import coral
from qt import *
from propertyeditors import *

from coral.modeler.property.propertyeditor import PrimitiveTypeEditorSingleLine
from coral.modeler import iconcache

L = coral.lang.CoralEx10.FileSystem
TM = coral.getTransactionManager()


class FSEntryQListItem(QListViewItem):
    def __init__(self, element,*args):
        
        self.name = unicode(element.name)
        self.element = element
        QListViewItem.__init__(self, *args)
        self.setText(0,unicode(self.name))
        if element.oclIsKindOf(L.File):
            self.setPixmap(0, QPixmap(iconcache.getIcon("filenewm")))
        else:
            self.setPixmap(0, QPixmap(iconcache.getIcon("directory")))


class ExtendedFolderEditor(FolderEditor):

    def __init__(self, editorSet, parent,manager=None):
        FolderEditor.__init__(self, editorSet, parent, manager)
        self.lstContents.connect(self.lstContents,SIGNAL("doubleClicked ( QListViewItem *, const QPoint &, int )"),self.openItem)
        self.lstContents.connect(self.lstContents,SIGNAL("selectionChanged( QListViewItem * )"),self.focusFS)
        self.lstContents.connect(self.lstContents,SIGNAL("clicked( QListViewItem * )"),self.deselectFS)
        self.delFSEntry.setIconSet(QIconSet(iconcache.getIcon("editdelete")))
        self.newFSEntry.setIconSet(QIconSet(iconcache.getIcon("linkAction")))
        self.selFolder.setIconSet(QIconSet(iconcache.getIcon("directory")))
        self.btnCommit.setIconSet(QIconSet(iconcache.getIcon("filesave")))
        
        ## Signals
        self.delFSEntry.setEnabled(False)
        self.btnCommit.setEnabled(False)
        self.selFolder.connect(self.selFolder,SIGNAL("clicked()"), self.selectFolder)
        self.btnCommit.connect(self.btnCommit,SIGNAL("clicked()"),self.commitEntry)

        self.CommitFolder.connect(self.CommitFolder,SIGNAL("textChanged( const QString &)"), self.change)
        
        self.delFSEntry.connect(self.delFSEntry, SIGNAL("clicked()"),self.deleteItem)        
        self.newFSEntry.connect(self.newFSEntry, SIGNAL("clicked()"),self.newItem)

    def change(self, qs):
        import string
        if string.strip(unicode(qs)):
            self.btnCommit.setEnabled(True)
        else:
            self.btnCommit.setEnabled(False)
        
    def focusOn(self,element):
        FolderEditor.focusOn(self,element)
        self.lstContents.clear()
        self.element = element
        for e in self.element.contents:
            i = FSEntryQListItem(e, self.lstContents)

    def openItem(self, item, *args):
        if item:
            e = item.element
            self.manager.focusOn(e)


    def newItem(self):
        TM.beginModifyCmd()
        try:
            prop=self.element.__type__().getPropertyByName("contents")
            asslist=AssociationListE(self.editorSet, self.parent(),prop,self.element,self.manager)
            asslist.show()
        finally:
            TM.endModifyCmd()         

    def deleteItem(self):
        sel = self.lstContents.selectedItem()
        if sel:
            TM.beginModifyCmd()
            try:
                sel.element.__resetAttrs__()
            finally:
                TM.endModifyCmd()
            self.delFSEntry.setEnabled(False)
            
    def focusFS(self, item):
        if item:
            self.delFSEntry.setEnabled(True)

    def deselectFS(self, item):
        if not item:
            self.delFSEntry.setEnabled(False)

        
    def selectFolder(self):
        qf = QFileDialog(self,"Select Commit Folder", True)

        folder = qf.getExistingDirectory()

        self.CommitFolder.setText(unicode(folder))

    def commitEntry(self):
        import os

        p = os.path.expanduser(unicode(self.CommitFolder.text()))

        try:
            self.element.commit(p)
        except:
            QMessageBox.critical(self,"Error","Unable to save files")


class ExtendedFileEditor(FileEditor):

    def __init__(self, editorSet, parent,manager=None):
        FileEditor.__init__(self, editorSet, parent, manager)
        self.selFolder.setIconSet(QIconSet(iconcache.getIcon("directory")))
        self.btnCommit.setIconSet(QIconSet(iconcache.getIcon("filesave")))
        self.btnPrint.setIconSet(QIconSet(iconcache.getIcon("printing")))
        self.ContentsEdit.setTextFormat(Qt.PlainText)
        
        ## Signals
        self.btnCommit.setEnabled(False)
        self.selFolder.connect(self.selFolder,SIGNAL("clicked()"), self.selectFolder)
        self.btnCommit.connect(self.btnCommit,SIGNAL("clicked()"),self.commitEntry)
        self.btnPrint.connect(self.btnPrint,SIGNAL("clicked()"),self.printDescription)
        self.CommitFolder.connect(self.CommitFolder,SIGNAL("textChanged( const QString &)"), self.change)

    def focusOn(self,element):
        FileEditor.focusOn(self,element)
        self.element = element
        if unicode(self.CommitFolder.text()):
            self.btnCommit.setEnabled(True)
        else:
            self.btnCommit.setEnabled(False)            
        
    def selectFolder(self):
        qf = QFileDialog(self,"Select Commit Folder", True)

        folder = qf.getExistingDirectory()

        self.CommitFolder.setText(unicode(folder))

    def commitEntry(self):
        import os

        p = os.path.expanduser(unicode(self.CommitFolder.text()))
        try:
            self.element.commit(p)
        except:
            QMessageBox.critical(self,"Error","Unable to save files")

    def change(self, qs):
        import string
        if string.strip(unicode(qs)):
            self.btnCommit.setEnabled(True)
        else:
            self.btnCommit.setEnabled(False)
        
    def printDescription(self):
        printer=QPrinter(QPrinter.HighResolution)
        printer.setFullPage(True)
        printer.setColorMode(QPrinter.Color)
        if printer.setup(self):
            p=QPainter(printer)
            if not p.isActive():
                return;
            metrics=QPaintDeviceMetrics(p.device())
            dpiy = metrics.logicalDpiY()
            margin = int(((2/2.54)*dpiy))
            body=QRect(margin, margin, metrics.width() - 2*margin, metrics.height()-2*margin)
            richText=QSimpleRichText( QStyleSheet.convertFromPlainText( self.ContentsEdit.text()),QFont(), self.ContentsEdit.context(),
                                      self.ContentsEdit.styleSheet(), self.ContentsEdit.mimeSourceFactory(),
                                      body.height() )
            richText.setWidth( p, body.width() )
            view=QRect(body)
            page = 1
            while True:
                richText.draw(p,body.left(),body.top(),view,self.colorGroup())
                view.moveBy(0,body.height())
                p.translate(0,-body.height())
                p.drawText(view.right()-p.fontMetrics().width(str(page)),
                           view.bottom()+p.fontMetrics().ascent()+5, str(page))
                p.drawText((view.left()+view.right()-p.fontMetrics().width(unicode(self.element.name)))/2,view.bottom()+p.fontMetrics().ascent()+5, unicode(self.element.name))
                if view.top()  >= richText.height():
                    break
                printer.newPage()
                page+=1
            p.end()
