import coral


class Proxy:
    """ A transformation library providing methods to simplify the templates """
    def __init__(self):
        pass

    def getParents(self, elem, me_included=False):
        """ Returns a string representing an element's parents' names separated with a dot, e.g. 'java.lang.String' """
        assert elem.oclIsKindOf(coral.lang.CORE01.AnyElement)

        temp = coral.Element__getParentsAndMyself__(elem, me_included)
        return ".".join([str(t.name) for t in temp])
        
    def getParentsOfType(self, type, me_included=False):
        """ Returns the type name if the passed type is a Java primitive datatype, otherwise same as getParents() """
        assert type.oclIsKindOf(coral.lang.CORE01.AnyElement)

        #if Java primitive datatype, don't return full path
        if type.name == "boolean":
            return type.name
        if type.name == "char":
            return type.name
        if type.name == "int":
            return type.name
        if type.name == "long":
            return type.name
        if type.name == "short":
            return type.name
        if type.name == "byte":
            return type.name
        if type.name == "float":
            return type.name
        if type.name == "double":
            return type.name
        
        temp = coral.Element__getParentsAndMyself__(type, me_included)
        return ".".join([str(t.name) for t in temp])

    
        


    def getImports(self,elem,elem_included=False):
        """ Returns a list of strings representing all the elements that a classifier is dependent on """
        L=coral.lang.UML14
        imports=[]
        found=False

        #Parameters
        for f in elem.feature:
            if f.oclIsKindOf(L.Operation):
                for p in f.parameter:
                    imp = self.getParents(p.type,elem_included)
                    imports.append(imp)
                    found = True

        #Associations
        for ae in elem.association.select(lambda l:l.oclIsKindOf(L.AssociationEnd)):
            for aend in ae.association.connection.select(lambda l:l != ae and l.isNavigable):
                imp = self.getParents(aend.participant,elem_included)
                imports.append(imp)
                found = True

        #Inheritance
        for g in elem.generalization:
            imp = self.getParents(g.parent,elem_included)
            imports.append(imp)
            found = True

        #Attributes
        for a in elem.feature:
            if a.oclIsKindOf(L.Attribute):
                imp = self.getParents(a.type,elem_included)
                imports.append(imp)
                found = True  

        if found:
            for i in imports[:]:
                if imports.count(i) > 1:
                    imports.remove(i)
        return imports


    def getReturnType(self, op):
        """ Returns a string representing the return type of an Operation """
        assert op.oclIsKindOf(coral.lang.UML14.Operation)
        for p in op.parameter:
            if p.kind == "return":
		return self.getParentsOfType(p.type,True)
        return "void"


    def getParameters(self, op):
        """ Returns a string representing the full paths of each parameter's type. The parameters are separated with commas.
        E.g. 'java.lang.String str, model.Vehicles.Wheel wheel' """
        assert op.oclIsKindOf(coral.lang.UML14.Operation)
        return ", ".join([str(self.getParentsOfType(p.type,True)+" "+p.name) for p in op.parameter if p.kind!="return"])
        

    def getSuperclasses(self, elem):
        """ Returns a string representing an element's superclasses' and implemented interfaces' names, separated with commas """
        return ", ".join([str(g.parent.name) for g in elem.generalization])

    def isExpressionMatch(self, match):
        """ Returns True if the parameter is an ExpressionMatch, False otherwise """
        if match.oclIsKindOf(coral.lang.MATCH10.ExpressionMatch):
            return True
        return False


    def isGraphMatch(self, match):
        """ Returns True if the parameter is a GraphMatch, False otherwise """
        if match.oclIsKindOf(coral.lang.MATCH10.GraphMatch):
            return True
        return False
    
