# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

import coral
DI = coral.lang.XMIDI20
from coral.modeler.extension import *
from coral.modeler import iconcache
from coral.modeler.diagram.degen import DiagramElementGenerator, DiagramElementCreatorExt
logger=coral.log.getLogger("UML14DiagramActions")

import coral.modeler.diagram.di_support as DISUPP

L = coral.lang.UML14
# The below import hack is not useful, use L instead

n = 0
def perpetualEnumerator():
    global n
    n += 1
    return n


class CreateAction( DiagramEditorToolbarEntry ):


    def prepare(self, diagram):
        self.__diagram = diagram

    def __call__(self, hints, parentge):
        return self.createElementAction(self.__diagram, self.__diagram, hints, parentge)

    def createElementAction( self, diagram, parent, hints=None, parentge=None ):
        if hints:
            hint1 = hints[0]
            if hint1:
                px = hint1[0]
                py = hint1[1]
            else:
                px = 50
                py = 50


        if parentge:
            model = parentge.semanticModel.element
        else:
            model = diagram.owner.element
        assert model
        
        coral.getTransactionManager().beginModifyCmd()
        try:
            element = self.createModelElement(model,parent)
            if not element:
                return
            deg = DiagramElementGenerator.getExtension(element.__type__())
            if not deg:
                return
            
            if model.oclIsKindOf(L.Package) and not model.oclIsKindOf(L.Model) and parentge:
                bc = None
                for cmp in parentge.contained:
                    if unicode(cmp.semanticModel.typeInfo) == "BodyCompartment":
                        bc = cmp
                        break
                if bc:
                    _x,_y = DISUPP.globalToLocalPosition(bc, px, py)
                    de = deg.createDiagramElement(element, xparent=bc, size=self.getDimension())
                    de.position.x = _x
                    de.position.y = _y
                    bc.contained.append(de)
            else:
                de = deg.createDiagramElement(element, xparent=parent, size=self.getDimension())
                if not de:
                    return
                de.position = DI.Point(x=px,y=py)
                parent.contained.append(de)
        finally:
            coral.getTransactionManager().endModifyCmd()

    def createModelElement( self, model, parent ):
        raise NotImplemented

    def createEdgeOrNot(self):
        return False

    def getDimension(self):
        raise NotImplemented

class CreateEdgeAction( CreateAction ):
    def firstType(self):
        raise NotImplemented

    def secondType(self):
        raise NotImplemented
    
    def createEdgeOrNot(self):
        return True

    def createElementAction(  self, diagram, parent, hints=None, parentge=None ):
        if not hints or not len(hints)==3:
            return
        model = diagram.owner.element
        assert model

        node1 = hints[0]
        node2 = hints[1]
        waypoints = hints[2]
        waypoints2 = []
        
        if ((node1.semanticModel and node1.semanticModel.element) and
            (node2.semanticModel and node2.semanticModel.element)):
            clsf1 = node1.semanticModel.element
            clsf2 = node2.semanticModel.element
            # BUG this checking should be more generic
            if not (clsf1.oclIsKindOf(self.firstType()) and clsf2.oclIsKindOf(self.secondType())):
                logger.debug("Attempt to make illegal edge")
                return
        else:
            logger.debug("Attempt to make illegal edge")
            return

        coral.getTransactionManager().beginModifyCmd()
        try:
            element = self.createModelElement(model,parent, node1, node2)
            if not element:
                return
            deg = DiagramElementGenerator.getExtension(element.__type__())
            if not deg:
                return
            anchor0 = DI.GraphConnector(graphElement=node1)
            anchor1 = DI.GraphConnector(graphElement=node2)
            ## Create the GraphEdge
            di_parent = DISUPP.getCommonParent(node1, node2)
            for wp in waypoints:
                ps_x, ps_y = coral.XMIDI_GlobalToLocalPosition(di_parent, wp.x, wp.y)
                waypoints2.append(DI.Point(x=ps_x, y=ps_y))
            de = deg.createDiagramElement(element, xparent = di_parent, waypoints=waypoints2 )
            assert de
            de.anchor.append(anchor0)
            de.anchor.append(anchor1)
            de.container = di_parent
            if waypoints:
                self.positionEdgeLabels(de, waypoints2)
        finally:
            coral.getTransactionManager().endModifyCmd()

    def createModelElement(self, model, parent, ge1, ge2):
        raise NotImplemented

    def positionEdgeLabels(self, graphEdge, waypoints):
        """Positions possible graphNodes below the @graphEdge, using the
        information in @waypoints. May be overriden by subclasses"""
        return

class CreateClassAction( CreateAction ):
    def getIcon( cls ):
        return iconcache.getIcon("icons/classtb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.Class(name="NewClass_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "Class"
    
    def getDimension(self):
        return DI.Dimension(width=100, height=100)

class CreatePackageAction( CreateAction ):
    def getIcon( cls ):
        return iconcache.getIcon("icons/packagetb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.Package(name="NewPackage_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "Package"
    
    def getDimension(self):
        return DI.Dimension(width=250, height=200)

class CreateInterfaceAction( CreateAction ):
    def createModelElement( self, model, parent ):
        c = L.Interface(name="NewInterface_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getIcon( cls ):
        return iconcache.getIcon("icons/interfacetb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Interface"
    
    def getDimension(self):
        return DI.Dimension(width=100, height=100)

class CreateAssociationAction( CreateEdgeAction ):
    def firstType(self):
        return L.Classifier

    def secondType(self):
        return L.Classifier

    def positionEdgeLabels(self, graphEdge, waypoints):


        assge = graphEdge
        assge.contained[2].position = DISUPP.calcGENameDIPoints(waypoints)
        
        de0p = assge.contained[0]
        de1p = assge.contained[1]
        wp1=waypoints[0]
        wp2=waypoints[-1]
        de0p.position.x, de0p.position.y = wp1.x, wp1.y
        de1p.position.x, de1p.position.y = wp2.x, wp2.y



        from coral.modeler.diagram.layout import _SizeHinter
        import coral.modeler.diagram.di_support_gfx as DISUPPGFX
        sh = _SizeHinter
        l =  GraphEdgeLayouterPureXMIDI(assge)

        l.layoutEdgeAndLabels(sh)

        ydiff = assge.waypoints[0].y - assge.waypoints[-1].y
        xdiff = assge.waypoints[0].x - assge.waypoints[-1].x
        
        if ydiff == 0:
            x, y = DISUPPGFX.calcTextSize(unicode(assge.semanticModel.element.name))
            assge.contained[2].position.x -= 0.5*x
        if xdiff == 0:
            assge.contained[2].position.x += 15

        if ydiff != 0:
            x, y = DISUPPGFX.calcTextSize(unicode(assge.semanticModel.element.name))
            assge.contained[2].position.x -= (x+5)
        

        for assend in [de0p,de1p]:

            part = assend.semanticModel.element.participant
            if not part:
                return
            partxmicand = part.xmidi_get_graphelements()
            if not partxmicand:
                return
            partxmi = None

            for a in assge.anchor:
                if a.graphElement in partxmicand:
                    partxmi = a.graphElement
                    break

            name = None
            vis = None
            mult = None
            for c in assend.contained:
                sm =  c.semanticModel
                if sm:
                    if unicode(sm.typeInfo) == 'Name':
                        name = c
                    if unicode(sm.typeInfo) == 'Visibility':
                        vis = c     
                    if unicode(sm.typeInfo) == 'Multiplicity':
                        mult = c

            if partxmi:
                px = float(partxmi.position.x)
                py = float(partxmi.position.y)
                pw = float(partxmi.size.width)
                ph = float(partxmi.size.height)

                ax = float(assend.position.x)
                ay = float(assend.position.y)


##                print bool(ax >= px+pw-1 and ay >= py and ay < ph+py), "right"
##                print ax, ">=", px+pw-1, "and", ay, ">=", py, "and", ay, "<", ph+py
##                print "\n"
##                print bool(ax <= px and (ay >= py and ay < ph+py)), "left"
##                print ax, "<=", px, "and", ay, ">=", py, "and", ay, "<", ph+py
##                print "\n"
##                print bool((ax >= px and ax < px+pw-1) and ay == py), "above"
##                print ax, ">=", px, "and", ax, "<", px+pw-1, "and", ay, "==", py
##                print "\n"
##                print bool(ax >= px and (ay >= py and ay >= ph+py-1)), "below"
##                print ax, ">=", px, "and", ay, ">=", py, "and", ay, ">=", ph+py-1
##                print "\n"
                
                # The association is fully on the right hand side of the class
                if ax >= px+pw-1 and (ay >= py and ay < ph+py):
                    if vis and name:
                        
                        vis.position.x = 15
                        vis.position.y = -20
                        name.position.x = vis.size.width + vis.position.x
                        name.position.y = vis.position.y
                    if not vis and name:
                        name.position.x = 15
                        name.position.y = -20
                    if mult:
                        mult.position.x = 15
                        mult.position.y = 10
                # The association is fully on the left hand side of the class
                elif ax <= px and (ay >= py and ay < ph+py):
                    if vis and name:
                        name_w, name_h = DISUPPGFX.calcTextSize(unicode(assend.semanticModel.element.name))
                        vis.position.x = -15 - vis.size.width - name_w
                        vis.position.y = -20
                        name.position.x = vis.size.width + vis.position.x
                        name.position.y = vis.position.y
                    if not vis and name:
                        if name.size.width < 20:
                            nsw = 40
                        else:
                            nsw = name.size.width
                        name.position.x = -15 - nsw
                        name.position.y = -20
                    if mult:
                        mult.position.x = -15 - mult.size.width
                        mult.position.y = 10
                # The association is fully above the class
                elif (ax >= px and ax < px+pw-1) and ay == py:
                    if vis and name:
                        name_w, name_h = DISUPPGFX.calcTextSize(unicode(assend.semanticModel.element.name))
                        vis.position.x = 15
                        vis.position.y = -15 - max(vis.size.height, name_h)
                        name.position.x = vis.size.width + vis.position.x
                        name.position.y = vis.position.y
                    if not vis and name:
                        name_w, name_h = DISUPPGFX.calcTextSize(unicode(assend.semanticModel.element.name))
                        name.position.x = 15
                        name.position.y = -15 - name_h
                    if mult:
                        mult.position.x = -15 - mult.size.width
                        mult.position.y = -15 - mult.size.height
                # The association is fully below the class
                elif ax >= px and (ay >= py and ay >= ph+py-1):
                    if vis and name:
                        vis.position.x = 15
                        vis.position.y = 15
                        name.position.x = vis.size.width + vis.position.x
                        name.position.y = vis.position.y
                    if not vis and name:
                        name.position.x = 15
                        name.position.y = 15
                    if mult:
                        mult.position.x = -15 - mult.size.width
                        mult.position.y = 15 
        
        

    def createModelElement(self, model, parent, ge1, ge2):
        ass = L.Association(name="NewAssociation_%d"%perpetualEnumerator())
        assend1 = L.AssociationEnd(name="the"+ ge1.semanticModel.element.name,association=ass, participant = ge1.semanticModel.element, isNavigable="true",multiplicity=L.Multiplicity(range=[L.MultiplicityRange(lower=1,upper=1)]))
        assend2 = L.AssociationEnd(name="the"+ ge2.semanticModel.element.name,association=ass, participant = ge2.semanticModel.element, isNavigable="true",multiplicity=L.Multiplicity(range=[L.MultiplicityRange(lower=1,upper=1)]))
        ass.connection.append(assend1)
        ass.connection.append(assend2)
        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(ass)
        return ass

    def getLabel(self):
        return "Association"
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/associationtb","UML14")
    getIcon = classmethod(getIcon)

        
class CreateGeneralizationAction( CreateEdgeAction ):
    def firstType(self):
        return L.GeneralizableElement

    def secondType(self):
        return L.GeneralizableElement

    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[0].position = DISUPP.calcGENameDIPoints(waypoints)

    def createModelElement(self, model, parent, ge1, ge2):
        me1 = ge1.semanticModel.element
        me2 = ge2.semanticModel.element

        if me1 == me2:
            # Don't allow self-inheritance
            # We can't prevent it totally, but this is one way less to do it
            logger.debug("Attempt to make illegal self inheritance")
            return

        gen = L.Generalization(name="NewGeneralization_%d"%perpetualEnumerator())
        gen.child = me1
        gen.parent = me2

        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(gen)
        return gen

    def getLabel(self):
        return "Generalization"
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/generalizationtb","UML14")
    getIcon = classmethod(getIcon)



class CreateDependencyAction( CreateEdgeAction ):
    def firstType(self):
        return L.ModelElement

    def secondType(self):
        return L.ModelElement

    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[0].position = DISUPP.calcGENameDIPoints(waypoints)

    def createModelElement(self, model, parent, ge1, ge2):
        # Create a binding if this seems realistic.
        C = L.Dependency
        if ge2.semanticModel.element.templateParameter:
            C = L.Binding
        dep = C(name="NewDependency_%d"%perpetualEnumerator())
        dep.client.append(ge1.semanticModel.element)
        dep.supplier.append(ge2.semanticModel.element)
        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(dep)
        return dep

    def getLabel(self):
        return "Dependency"
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/dependencytb","UML14")
    getIcon = classmethod(getIcon)

class CreateCommentLinkAction( CreateAction ):
    def createElementAction(  self, diagram, parent, hints=None, parentge=None ):
        if hints and len(hints)==3:
            model = diagram.owner.element
            assert model
        node1 = hints[0]
        node2 = hints[1]
        waypoints = hints[2]
        waypoints2 = []
        if ((node1.semanticModel and node1.semanticModel.element) and
            (node2.semanticModel and node2.semanticModel.element)):
            me1 = node1.semanticModel.element
            me2 = node2.semanticModel.element
            if not (me1.oclIsKindOf(L.Comment) ^ me2.oclIsKindOf(L.Comment)):
                logger.debug("Attempt to make illegal comment link")
                return
        else:
            logger.debug("Attempt to make illegal comment link")
            return

        if me1.oclIsKindOf(L.Comment):
            me2.comment.append(me1)
        else:
            me1.comment.append(me2) 

        coral.getTransactionManager().beginModifyCmd()
        try:
            di_parent = DISUPP.getCommonParent(node1, node2)
            for wp in waypoints:
                ps_x, ps_y = coral.XMIDI_GlobalToLocalPosition(di_parent, wp.x, wp.y)
                waypoints2.append(DI.Point(x=ps_x, y=ps_y))            
            de = DiagramElementCreatorExt().createSimpleGraphEdge("CommentLink", waypoints=waypoints2)
            if de:
                anchor0 = DI.GraphConnector(graphElement=node1, position=DI.Point())
                anchor1 = DI.GraphConnector(graphElement=node2, position=DI.Point())
                de.anchor.append(anchor0)
                de.anchor.append(anchor1)
                de.container = di_parent
        finally:
            coral.getTransactionManager().endModifyCmd()

    def getLabel(self):
        return "Comment Link"
        

    def createModelElement( self, model, parent ):
        pass

    def getIcon( cls ):
        return iconcache.getIcon("icons/commentlinktb","UML14")
    getIcon = classmethod(getIcon)

    def createEdgeOrNot(self):
        return True


class CreateActorAction( CreateAction ):

    def createModelElement( self, model, parent ):
        a = L.Actor(name="NewActor_%d"%perpetualEnumerator())
        model.ownedElement.insert(a)
        return a

    def getIcon( cls ):
        return iconcache.getIcon("icons/actortb","UML14")
    getIcon = classmethod(getIcon)


    def getLabel(self):
        return "Actor"
    
    def getDimension(self):
        return DI.Dimension(width=70, height=120)


class CreateSubsystemAction( CreateAction ):

    def getIcon( cls ):
        return iconcache.getIcon("icons/subsystemtb","UML14")
    getIcon = classmethod(getIcon)


    def getLabel(self):
        return "Subsystem"
    

class CreateComponentAction( CreateAction ):

    def createModelElement( self, model, parent ):
        c = L.Component(name="NewComponent_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/componenttb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Component"
    
    def getDimension(self):
        return DI.Dimension(width=100, height=100)

class CreateCommentAction( CreateAction ):

    def createModelElement( self, model, parent ):
        c = L.Comment(name="NewComment_%d"%perpetualEnumerator())
        from coral import configuration
        import time
        user = configuration.getUserName()
        gmt = time.gmtime()
        c.body.__setString__(
            "\nComment made by %s on %d-%02d-%02d at %02d:%02d:%02d (UTC)"
            % (user, gmt[0], gmt[1], gmt[2],
               gmt[3], gmt[4], gmt[5]))
        # Don't set .comment for model, because
        # that's not really what we want. When the user places
        # a CommentLink, we know what to connect
        
        # Add the comment to the first .ownedElement we can
        while not model.__type__().getPropertyByName("ownedElement"):
            model = model.__parent__()
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "Comment"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/commenttb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=100, height=100)


##=========================================
##=========================================
##=========State=Diagrams==================
##=========================================
##=========================================
class CreateStateAction( CreateAction ):
    def createElementAction( self, diagram, parent, hints=None, parentge=None ):
        if hints:
            hint1 = hints[0]
            if hint1:
                px = hint1[0]
                py = hint1[1]
            else:
                px = 50
                py = 50

        if parentge and parentge.semanticModel.element.oclIsKindOf(L.CompositeState):
            model = parentge.semanticModel.element
        else:
            model = diagram.owner.element        

        assert model
        
        coral.getTransactionManager().beginModifyCmd()
        try:
            element = self.createModelElement(model,parent)
            if not element:
                return
            deg = DiagramElementGenerator.getExtension(element.__type__())
            if not deg:
                return
            
            if model.oclIsKindOf(L.StateMachine):
                de = deg.createDiagramElement(element, xparent=parent, size=self.getDimension())
                if not de:
                    return
                de.position = DI.Point(x=px,y=py)
                parent.contained.append(de)
            else:
                ssc = None
                for cmp in parentge.contained:
                    if not cmp.semanticModel.oclIsKindOf(DI.Uml1SemanticModelBridge) and unicode(cmp.semanticModel.typeInfo) == "SubStateCompartment":
                        ssc = cmp
                        break
                if ssc:
                    _x,_y = DISUPP.globalToLocalPosition(ssc, px, py)
                    de = deg.createDiagramElement(element, xparent=ssc, size=self.getDimension())
                    if not de:
                        return
                    de.position = DI.Point(x=_x, y=_y)
                    ssc.contained.append(de)
                elif not ssc and model.oclIsKindOf(L.CompositeState):
                    _x,_y = DISUPP.globalToLocalPosition(parentge, px, py)
                    de = deg.createDiagramElement(element, xparent=parentge, size=self.getDimension())
                    if not de:
                        return
                    de.position = DI.Point(x=_x, y=_y)
                    parentge.contained.append(de)
        finally:
            coral.getTransactionManager().endModifyCmd()



class CreateSimpleStateAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.SimpleState(name="NewSimpleState_%d"%perpetualEnumerator())
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "SimpleState"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/statetb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=100, height=100)


class CreateCompositeStateAction( CreateStateAction ):
    def createModelElement( self, model, parent ):
        c = L.CompositeState(name="NewCompositeState_%d"%perpetualEnumerator())
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "CompositeState"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/compstatetb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=100, height=100)


class CreateConcurrentStateAction( CreateStateAction ):
    def createModelElement( self, model, parent ):
        c = L.CompositeState(name="NewConcurrentState_%d"%perpetualEnumerator(), isConcurrent=True)
        is1 = L.CompositeState(name="Region_1", container=c)
        is2 = L.CompositeState(name="Region_2", container=c)
        c.subvertex.insert(is1)
        c.subvertex.insert(is2)
        
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "Concurrent CompositeState"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/cncstatetb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=100, height=150)


class CreateSynchStateAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.SynchState(name="NewSynchState_%d"%perpetualEnumerator())
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/synchtb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=20, height=20)

    def getLabel(self):
        return "SynchState"
    

class CreateFinalStateAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.FinalState(name="NewFinalState_%d"%perpetualEnumerator())
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/finalstatetb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=20, height=20)

    def getLabel(self):
        return "FinalState"
    

class CreateTransitionAction( CreateEdgeAction ):
    def firstType(self):
        return L.StateVertex

    def secondType(self):
        return L.StateVertex

    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[1].position = DISUPP.calcGENameDIPoints(waypoints)

    def createModelElement( self, model, parent, ge1, ge2 ):
        c = L.Transition(name="NewTransition_%d"%perpetualEnumerator())
        c.source = ge1.semanticModel.element
        c.target = ge2.semanticModel.element
        model.transitions.insert(c)
        return c

    def getLabel(self):
        return "Transition"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/transitiontb","UML14")
    getIcon = classmethod(getIcon)

class CreateForkAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewFork_%d"%perpetualEnumerator(), kind="fork")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "Fork"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/forktb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=50, height=10)

class CreateInitialStateAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewInitialState_%d"%perpetualEnumerator(), kind="initial")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/initialstatetb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Initial State"
    
    def getDimension(self):
        return DI.Dimension(width=20, height=20)
    
class CreateJoinAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewJoin_%d"%perpetualEnumerator(), kind="join")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/jointb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Join"
    
    def getDimension(self):
        return DI.Dimension(width=50, height=10)

class CreateJunctionAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewJunction_%d"%perpetualEnumerator(), kind="junction")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "Junction"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/junctiontb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=20, height=20)


class CreateChoiceAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewChoice_%d"%perpetualEnumerator(), kind="choice")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/choicetb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Choice"
    
    def getDimension(self):
        return DI.Dimension(width=20, height=20)

class CreateDeepHistoryAction( CreateStateAction ):

    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewDeepHistory_%d"%perpetualEnumerator(), kind="deepHistory")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/deephistorytb","UML14")
    getIcon = classmethod(getIcon)

    def getLabel(self):
        return "Deep History"
    
    def getDimension(self):
        return DI.Dimension(width=20, height=20)


class CreateShallowHistoryAction( CreateStateAction ):

    
    def createModelElement( self, model, parent ):
        c = L.Pseudostate(name="NewShallowHistory_%d"%perpetualEnumerator(), kind="shallowHistory")
        if model.oclIsKindOf(L.StateMachine):
            model.top.subvertex.insert(c)
        elif model.oclIsKindOf(L.CompositeState):
            model.subvertex.insert(c)
        return c

    def getLabel(self):
        return "Shallow History"
        
    def getIcon( cls ):
        return iconcache.getIcon("icons/historytb","UML14")
    getIcon = classmethod(getIcon)

    def getDimension(self):
        return DI.Dimension(width=20, height=20)

## Object Diagrams and Collaboration Diagrams


class CreateObjectAction( CreateAction ):

    def getIcon( cls ):
        return iconcache.getIcon("icons/objecttb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.Object(name="NewObject_%d"%perpetualEnumerator())

        if model.oclIsKindOf(L.CollaborationInstanceSet):
            model.collaboration.ownedElement.insert(c)
            model.participatingInstance.insert(c)
        else:
            model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "Object"
    
    def getDimension(self):
        return DI.Dimension(width=120, height=80)


class CreateLinkAction( CreateEdgeAction ):
    def firstType(self):
        return L.Instance

    def secondType(self):
        return L.Instance

    def createModelElement(self, model, parent, ge1, ge2):
         lnk = L.Link(name="NewLink_%d"%perpetualEnumerator())
         lnkend1 = L.LinkEnd(link=lnk, instance = ge1.semanticModel.element)
         lnkend2 = L.LinkEnd(link=lnk, instance = ge2.semanticModel.element)
         lnk.connection.append(lnkend1)
         lnk.connection.append(lnkend2)
         lnk.owner_at_Link = ge1.semanticModel.element
         if model.oclIsKindOf(L.CollaborationInstanceSet):
             model.participatingLink.insert(lnk)
         return lnk

    def getLabel(self):
        return "Link"
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/linktb","UML14")
    getIcon = classmethod(getIcon)

# Deployment and Component Diagrams

class CreateNodeAction( CreateAction ):

    def getIcon( cls ):
        return iconcache.getIcon("icons/nodetb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.Node(name="NewNode_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getDimension(self):
        return DI.Dimension(width=120, height=120)

    def getLabel(self):
        return "Node"
    

class CreateNodeInstanceAction( CreateAction ):

    def getIcon( cls ):
        return iconcache.getIcon("icons/nodeinstancetb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.NodeInstance(name="NewNodeInstance_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getDimension(self):
        return DI.Dimension(width=120, height=120)

    def getLabel(self):
        return "NodeInstance"
    
class CreateComponentInstanceAction( CreateAction ):

    def getIcon( cls ):
        return iconcache.getIcon("icons/componentinstancetb","UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.ComponentInstance(name="NewComponentInstance_%d"%perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "ComponentInstance"
    
    def getDimension(self):
        return DI.Dimension(width=120, height=120)

class CreateAbstractionAction( CreateEdgeAction ):
    def firstType(self):
        return L.ModelElement

    def secondType(self):
        return L.ModelElement

    def createModelElement(self, model, parent, ge1, ge2):
        abs = L.Abstraction(name="NewAbstraction_%d"%perpetualEnumerator())
        abs.client.append(ge1.semanticModel.element)
        abs.supplier.append(ge2.semanticModel.element)
        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(abs)
        return abs

    def getLabel(self):
        return "Abstraction"
    
    def getIcon( cls ):
        return iconcache.getIcon("icons/abstractiontb","UML14")
    getIcon = classmethod(getIcon)

    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[0].position = DISUPP.calcGENameDIPoints(waypoints)


#
#
# USE CASE DIAGRAMS
#
#

class CreateUseCaseAction(CreateAction):
    def getIcon( cls ):
        return iconcache.getIcon("icons/usecasetb", "UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.UseCase(name="New Use Case %d" % perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "UseCase"
    
    def getDimension(self):
        return DI.Dimension(width=120, height=80)

class CreateCollaborationAction(CreateAction):
    def getIcon( cls ):
        return iconcache.getIcon("icons/collaborationtb", "UML14")
    getIcon = classmethod(getIcon)

    def createModelElement( self, model, parent ):
        c = L.Collaboration(name="New Collaboration %d" % perpetualEnumerator())
        model.ownedElement.insert(c)
        return c

    def getLabel(self):
        return "Collaboration"
    
    def getDimension(self):
        return DI.Dimension(width=120, height=80)

class CreateExtendAction( CreateEdgeAction ):
    def firstType(self):
        return L.UseCase

    def secondType(self):
        return L.UseCase

    def createModelElement(self, model, parent, ge1, ge2):
        extend = L.Extend(name="Extend %d"%perpetualEnumerator())
        extend.base = ge1.semanticModel.element
        extend.extension = ge2.semanticModel.element
        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(extend)         
        return extend

    def getLabel(self):
        return "Extend"
    
    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[0].position = DISUPP.calcGENameDIPoints(waypoints)

    def getIcon( cls ):
        return iconcache.getIcon("icons/extendtb","UML14")
    getIcon = classmethod(getIcon)

class CreateIncludeAction( CreateEdgeAction ):
    def firstType(self):
        return L.UseCase

    def secondType(self):
        return L.UseCase

    def createModelElement(self, model, parent, ge1, ge2):
        include = L.Include(name="Include %d" % perpetualEnumerator())
        include.base = ge1.semanticModel.element
        include.addition = ge2.semanticModel.element
        abs_parent = DISUPP.getCommonParent(ge1.semanticModel.element, ge2.semanticModel.element)
        if not abs_parent:
            di_parent = DISUPP.getCommonParent(ge1, ge2)
            abs_parent = DISUPP.findUpwardsModelElement(di_parent)
        abs_parent.ownedElement.add(include)         
        return include

    def getLabel(self):
        return "Include"
    
    def positionEdgeLabels(self, graphEdge, waypoints):
        graphEdge.contained[0].position = DISUPP.calcGENameDIPoints(waypoints)

    def getIcon( cls ):
        return iconcache.getIcon("icons/includetb","UML14")
    getIcon = classmethod(getIcon)





### NOTE: This is a hacked graphedge layouter, used to layout an edge
### before a widget has been created. DO NOT USE, incompatible with other layouters!
from coral.modeler.diagram.layout import connectorToEdgeCoordinates,floatEquals, _floatEquals, clamp
class GraphEdgeLayouterPureXMIDI:
    def __init__( self, ge ):
        self.graphElement = ge

    def layout(self, sizeHinter,
               toMinimal=False, w_parent = -1, h_parent = -1 ):
        """Default layout rearranges the waypoints."""
        self.moveContained(sizeHinter)

    def move(self, x, y):
        """Edges cannot be moved."""
        pass

    def resize(self, w, h):
        """Edges don't need to be resized."""
        pass

    def getSize(self):
        """Edges don't have a size."""
        return 0.0, 0.0

    def sizeHint(self, sizeHinter, toMinimal):
        """Override sizeHint for edges. We don't have a size.
        I wonder if this is bad or surprising behavior?"""
        return 0, 0


    def layoutEdgeAndLabels(self, sizeHinter,
                            connectstart=True, connectend=True ):
        # The waypoints of an edge are relative to the container
        
        # If an edge connects elements from different containers the
        # edge is relative to the container containing the both containers:-).
        
        # If edges are only moved relatively this is not a problem but if we have
        # to calculate waypoint values from scratch the common container has to be known.
        graphEdge=self.graphElement

        if graphEdge.oclIsKindOf(DI.GraphNode):
            return

        if not len(graphEdge.anchor) == 2:
            print "BUG: Attempt to layout a graphEdge (",graphEdge, graphEdge.getName(), graphEdge.semanticModel.element,") with",len(graphEdge.anchor) ,"anchors"
            return

        startconnector=graphEdge.anchor[0]
        endconnector=graphEdge.anchor[1]

        # Find the moved distance by comparing the moved connectors
        # to the waypoints that have not moved
        sX,sY=connectorToEdgeCoordinates(startconnector, graphEdge)
        eX,eY=connectorToEdgeCoordinates(endconnector, graphEdge)
        wsX,wsY=graphEdge.waypoints[0].x,graphEdge.waypoints[0].y
        weX,weY=graphEdge.waypoints[-1].x,graphEdge.waypoints[-1].y
        startDiff=(sX-wsX,sY-wsY)
        endDiff=(eX-weX,eY-weY)

        # If both connectors have moved the same distance then we translate
        # or if it is a self association, only start==end doesn't work.
        #if startconnector.graphElement==endconnector.graphElement:
        #    self.__handleSelfAssociations(startconnector, endconnector)
        if floatEquals(startDiff,endDiff) and not floatEquals(startDiff,(0.0,0.0)):
            self.__translate(startconnector, endconnector)
        else:
            # Move connectors somehow to make the lines straight
            self.slideConnector(startconnector,graphEdge)
            self.slideConnector(endconnector,graphEdge)

            # Connect end points to connectors
            X,Y=connectorToEdgeCoordinates(startconnector, graphEdge)
            if connectstart:
                startDiff=(X-graphEdge.waypoints[0].x,Y-graphEdge.waypoints[0].y)
                graphEdge.waypoints[0].x=X
                graphEdge.waypoints[0].y=Y
            else:
                startDiff = 0.0,0.0

            X,Y=connectorToEdgeCoordinates(endconnector, graphEdge)
            endDiff=(X-graphEdge.waypoints[-1].x,Y-graphEdge.waypoints[-1].y)
            if connectend:
                graphEdge.waypoints[-1].x=X
                graphEdge.waypoints[-1].y=Y
            else:
                endDiff = 0.0,0.0

            self.moveLabels(sizeHinter, startDiff, endDiff)


    def moveContained(self, sizeHinter):
        self.layoutEdgeAndLabels(sizeHinter, True,True)
        

    def __translate(self, startconnector, endconnector):
        elem=startconnector.graphElement
        edge=self.graphElement

        # Find difference between connector and waypoint coordinates
        dx,dy=connectorToEdgeCoordinates(startconnector, edge)
        dx-=edge.waypoints[0].x
        dy-=edge.waypoints[0].y

        # Translate all waypoints by the same offset
        for p in edge.waypoints:
            p.x+=dx
            p.y+=dy
            
        # Translate labels
        for gnode in edge.contained:
            gnode.position.x+=dx
            gnode.position.y+=dy

        # Translate owned graph connectors
        for graphconnector in edge.anchorage:
            graphconnector.position.x+=dx
            graphconnector.position.y+=dy
            

    def moveLabels(self, sizeHinter, startDiff, endDiff):
        """
        If the ends have moved different distances then try to move
        labels relative to their ends..
        """
        edge=self.graphElement

        if len(edge.contained) == 3:
            e0p = edge.contained[0].position
            e0p.x += startDiff[0]
            e0p.y += startDiff[1]
            
            e1p = edge.contained[1].position
            e1p.x += endDiff[0]
            e1p.y += endDiff[1]

            if len(edge.waypoints) == 2:
                middleDiff = ((startDiff[0]+endDiff[0]) / 2,
                              (startDiff[1]+endDiff[1]) / 2)
                e2p = edge.contained[2].position
                e2p.x += middleDiff[0]
                e2p.y += middleDiff[1]
            

    def slideConnector( self, con, edge ):
        assert len(edge.anchor)==2

        if edge.anchor[0]==con:
            firstwp = edge.waypoints[0]
            secondwp = edge.waypoints[1]
            oppositecon = edge.anchor[1]
        elif edge.anchor[1]==con:
            firstwp = edge.waypoints[-1]
            secondwp = edge.waypoints[-2]
            oppositecon = edge.anchor[0]
        else:
            raise "Can't happen :-( ???"

        # goal point, i.e. the point to which the distance from the connector
        # should be minimized
        # set goal to the position of the second waypoint
        goalx,goaly = coral.XMIDI_LocalToGlobalPosition(edge,float(secondwp.x),float(secondwp.y))

        # special case for edges with 2 waypoints if the opposite connector
        # is a GraphNode
        if len(edge.waypoints)==2 and oppositecon.graphElement.oclIsKindOf(DI.GraphNode):
            # yes, I know this is a bit fugly

            # get the rectangular bounds of the opposite GraphNode
            onode = oppositecon.graphElement
            oleft,otop = coral.XMIDI_LocalToGlobalPosition(onode,0.0,0.0)
            oright = oleft+float(onode.size.width)-1.0
            obottom = otop+float(onode.size.height)-1.0

            # if we have a GraphNode to GraphNode edge
            if con.graphElement.oclIsKindOf(DI.GraphNode):
                # coordinates of first waypoint (the one at con)
                fwpx,fwpy = coral.XMIDI_LocalToGlobalPosition(edge,float(firstwp.x),float(firstwp.y))

                # if the first waypoint is inside in the rectangular bounds of
                # the opposite GraphNode, we have an "inside" edge
                if oleft<fwpx<oright and otop<fwpy<obottom:
                    node = con.graphElement
                    left,top = coral.XMIDI_LocalToGlobalPosition(node,0.0,0.0)
                    right = left+float(node.size.width)-1.0
                    bottom = top+float(node.size.height)-1.0

                    segments = self.__getSegmentsForConnector(oppositecon)
                    fwpx = clamp(fwpx,left,right)
                    fwpy = clamp(fwpy,top,bottom)

                    # set the goal point to the goal point that would have
                    # been chosen from the other connector end, this ensures
                    # correct snapping to the inside of a GraphNode
                    goalx,goaly = self.__getNewConnectorPosition(segments,fwpx,fwpy)
                else:
                    # else, set goal to the midpoint of the two waypoints
                    goalx = (float(firstwp.x) + float(secondwp.x) ) / 2.0
                    goaly = (float(firstwp.y) + float(secondwp.y) ) / 2.0
                    goalx,goaly = coral.XMIDI_LocalToGlobalPosition(edge,goalx,goaly)

            # ensure goalx, goaly are within rectangular bounds of the
            # opposite GraphNode
            goalx = clamp(goalx,oleft,oright)
            goaly = clamp(goaly,otop,obottom)

        elif len(edge.waypoints)==2 and oppositecon.graphElement.oclIsKindOf(DI.GraphEdge):
            goalx = (float(firstwp.x) + float(secondwp.x) ) / 2.0
            goaly = (float(firstwp.y) + float(secondwp.y) ) / 2.0
            goalx,goaly = coral.XMIDI_LocalToGlobalPosition(edge,goalx,goaly)

        segments = self.__getSegmentsForConnector(con)
        conx,cony = self.__getNewConnectorPosition(segments,goalx,goaly)

        con.position.x,con.position.y = coral.XMIDI_GlobalToLocalPosition(con.graphElement,conx,cony)
   
        
    def __getNewConnectorPosition( self, segments, goalx, goaly ):
        orthopoints = []
        otherpoints = []
        for (x1,y1,x2,y2) in segments:
            if min(x1,x2)<=goalx<=max(x1,x2) and abs(x1-x2)>0.1:
                x = goalx
                y = y1 + (y2-y1) * (x-x1) / (x2-x1)
                orthopoints.append((x,y))
            elif min(y1,y2)<=goaly<=max(y1,y2) and abs(y1-y2)>0.1:
                y = goaly
                x = x1 + (x2-x1) * (y-y1) / (y2-y1)
                orthopoints.append((x,y))
            else:
                otherpoints.append( ((x1+x2)/2,(y1+y2)/2) )
                otherpoints.append( (x1,y1) )
                otherpoints.append( (x2,y2) )

        # choose the closest orthopoint, or if none, the closest
        # otherpoint
        closest = None
        cx,cy = 0.0, 0.0
        for pts in orthopoints,otherpoints:
            for x,y in pts:
                dist = (x-goalx)**2 + (y-goaly)**2
                if closest == None or dist < closest:
                    closest = dist
                    cx,cy = x,y
            if closest != None:
                return cx,cy
        return cx,cy
        

    def __getSegmentsForConnector( self, con ):
        """
        Returns a list of line segments (x1,y1,x2,y2) that decribes the
        edges along which the connector con may slide. Currently it returns
        four edges for DI.GraphNodes, and the segments between waypoints for
        DI.GraphEdges. All coordinates should be global.
        
        Since this function returns arbitrary line segments, we could
        make an extension that supports GraphNodes with shapes. This
        could be useful for laying out edges to non-rectangular GraphNodes
        such as UML use cases and packages.
        """
        if con.graphElement.oclIsKindOf(DI.GraphNode):
            node = con.graphElement
            x1,y1 = coral.XMIDI_LocalToGlobalPosition(node,0.0,0.0)
            x2 = x1 + float(node.size.width)-1.0
            y2 = y1 + float(node.size.height)-1.0
            segments = [ (x1,y1,x2,y1),
                         (x1,y1,x1,y2),
                         (x2,y1,x2,y2),
                         (x1,y2,x2,y2) ]
        elif con.graphElement.oclIsKindOf(DI.GraphEdge):
            edge = con.graphElement
            xofs,yofs = coral.XMIDI_LocalToGlobalPosition(edge,0.0,0.0)
            segments = []
            for i in xrange(len(edge.waypoints)-1):
                x1 = float(edge.waypoints[i].x) + xofs
                y1 = float(edge.waypoints[i].y) + yofs
                x2 = float(edge.waypoints[i+1].x) + xofs
                y2 = float(edge.waypoints[i+1].y) + yofs
                segments.append((x1,y1,x2,y2))
        return segments


