#!/usr/bin/env python
#
# uml2any - Simple UML 1.4 to Any languager converter infrastructure
#
# GOALS
#
# Converts class diagrams into java packages and classes
# The operations are "dummy", writing out what method was called.
#
# Converts collaboration diagrams
# This creates a main() method which creates the initial objects
# and does some kind of method calls on them. We'll see...
#
# AUTHORS
#
# Marcus Alanen 2005-06
#
#
#
import sys
import coral
import os

import code_generator

from coral.lgen.codeartifact import CodeArtifact
L = coral.meta.UML14

def convert(cgen, models, outputdir):
    """@cgen is a Python module with helper routines.
    @models is any collection of UML 1.4 Models that should
    be converted."""
    if type(models) == str:
        models = coral.loadDocument(models).contents()

    code_generator.ensure_dirs( [ outputdir ] )
    os.chdir(outputdir)
    
    F_elems = {}
    ca_makefile = CodeArtifact()
    
    F_elems[0] = ("makefile", [ ca_makefile ] )
    
    all = coral.Set()
    for root in models:
        if root.oclIsKindOf(L.root_ed):
            all.unionWith(root.__getAllParts__())
    
    objects = all.select(lambda x: x.oclIsKindOf(L.Object))
    cgen.create_main.create_main_class(all, objects)

    cgen.fix_uml_model.fix(all)
    
    ### Ensure directories exist #############################################
    
    def create_hierarchy(e):
        l = coral.Element__getParentsAndMyself__(e)
        code_generator.ensure_dirs( [ str(i.name) for i in l ] )
    
    for e in all.select(lambda x: x.oclIsKindOf(L.Package)):
        create_hierarchy(e)
    
    for e in all:
        if e.oclIsKindOf(L.Classifier):
            cgen.create_static_class_structure.create_class(F_elems, e)
    
    ### OUTPUT EVERYTHING ################################################
    
    for k,v in F_elems.iteritems():
        fname, cas = v
        f = file(fname, "w")
        try:
            for ca in cas:
                ca.dump_fd(f)
        finally:
            f.close()

if __name__ == "__main__":
    f = sys.argv[1]
    if len(sys.argv) > 2:
        outputdir = sys.argv[2]
    else:
        outputdir = "tmp"
    prj = coral.loadDocument(f)
    convert(None, prj.contents(), outputdir)
