import coral
import jhelp
import os

L = coral.meta.UML14

from coral.lgen.codeartifact import CodeArtifact

### METHOD ###########################################################

def _get_types(e):
    k = []
    internals = []
    k.append(e)
    for i in e.clientDependency:
        found_internal = False
        for s in i.stereotype:
            if s.name == "internal":
                found_internal = True
        if i.oclIsKindOf(L.Binding):
            for a in i.argument:
                if found_internal:
                    internals.append(a.modelElement)
                else:
                    k.append(a.modelElement)
        for s in i.supplier:
            assert s
            if found_internal:
                internals.append(s)
            else:
                k.append(s)
    return k, internals

def create_method(method, c, ch):
    ch.ca.add("")
    t_static = ""
    if method.specification.ownerScope == "instance":
        pass
    elif method.specification.ownerScope == "classifier":
        t_static = "static "
    else:
        assert 0

    res = method.specification.name
    res += "("
    # Loop through method formal parameters
    params = []
    retparams = []
    for param in method.specification.parameter:
        if param.kind == "return":
            retparams.append(param)
            continue
        params.append(param)

    def f(params):
        l = []
        l_no_default = []
        for p in params:
            assert p.type
            l1,l2 = _get_types(p.type)
            for i in l1:
                c.import_these.add(i)
                ch.import_these.add(i)
            for i in l2:
                c.import_these.add(i)

            s = jhelp.sep_name(p.type, '::') + " " + str(p.name)
            l_no_default.append(s)
            if p.defaultValue and p.defaultValue.body:
                s += " = " + p.defaultValue.body
            l.append(s)
        return ", ".join( l ), ", ".join( l_no_default )
    
    # BUG loop through keywords args
    rp, rp_no_default = f(retparams)
    assert rp == rp_no_default
    if not rp:
        rp += jhelp.sep_name(jhelp.cpp.lang.void, '::')
    rp += " "
    rf, rf_no_default = f(params)
    resc = rp + str(method.owner.name) + "::" + res + rf_no_default

    resc += ")"
    ch.ca.add(jhelp.visibility_u2j(method.specification.visibility,
                                   whitespace=False) + ":")
    ch.ca.indent(1)
    if method.specification.ownerScope != "classifier" \
           and ( method.specification.isAbstract or not method.specification.isLeaf ):
        rvirt = "virtual "
    else:
        rvirt = ""

    if method.specification.isQuery:
        rconst = " const"
    else:
        rconst = ""
    resh = t_static + rvirt + rp + res + rf + ")" + rconst
    resc += rconst
    # COMMENTS
    l = []
    l.extend(method.specification.comment.body)
    l.extend(method.comment.body)
    for param in method.specification.parameter:
        if param.comment:
            s = []
            s.extend(param.comment.body)
            s[0] = "@" + unicode(param.name) + " - " + unicode(s[0])
            l.extend( s )
        
    jhelp.add_comment(ch.ca, l)
    
    ch.ca.add(resh + ";")

    c.ca.add(resc + " {")
    # Method body
    c.ca.indent(1)
    b = ""
    if method.body:
        b = unicode(method.body.body).strip()

    if b == "" and method.specification.specification:
        b = unicode(method.specification.specification)

    if b == "":
        c.ca.add("/* No method body defined */")
        c.ca.add("std::clog << \"" + resc + "\" << std::endl;")
        c.import_these.add(jhelp.std.iostream)
    else:
        b = b.split("\n")
        for i in b:
            c.ca.add(i)
    c.ca.indent(-1)
    c.ca.add("}")
    ch.ca.indent(-1)

### ATTRIBUTES ###########################################################

def create_attribute(ch, name, type, multiplicity, ordering, visibility, comment):
    l = 0
    u = 1
    if multiplicity:
        for r in multiplicity.range:
            l = min(l, r.lower)
            u = max(u, r.upper)

    ch.import_these.add(type)
    if u > 1:
        coltype = None
        if ordering == "ordered":
            coltype = jhelp.std.vector
        else:
            coltype = jhelp.std.set

        type = jhelp._instantiate_template(coltype, [ type ])

    assert type
    l1,l2 = _get_types(type)
    for i in l1:
        ch.import_these.add(i)

    ch.ca.add(jhelp.visibility_u2j(visibility, whitespace=False) + ":")
    ch.ca.indent(1)
    jhelp.add_comment(ch.ca, comment)
    ch.ca.add(jhelp.sep_name(type, '::') + " " + name + ";")
    ch.ca.indent(-1)


### CLASS/INTERFACE ######################################################

def create_class(F_elems, e):
    ch = jhelp.Creator()
    ch.import_these = coral.Set()
    ch.ca_header = CodeArtifact()
    ch.ca_imports = CodeArtifact()
    ch.ca_ns_decl = CodeArtifact()
    ch.ca = CodeArtifact()
    ch.ca_ns_end = CodeArtifact()
    ch.ca_footer = CodeArtifact()
    ch.ca_list = [ ch.ca_header, ch.ca_imports,
                   ch.ca_ns_decl, ch.ca, ch.ca_ns_end, ch.ca_footer ]
    
    pkglist = coral.Element__getParentsAndMyself__(e, False)

    pkgnames = [ str(i.name) for i in pkglist ]
    
    c = jhelp.Creator()
    c.import_these = coral.Set()
    c.import_these.add(e) # Always import our own .h file
    c.ca_header = CodeArtifact()
    c.ca_imports = CodeArtifact()
    c.ca_ns_decl = CodeArtifact()
    c.ca = CodeArtifact()
    c.ca_footer = CodeArtifact()
    c.ca_list = [ c.ca_header, c.ca_imports, c.ca_ns_decl, c.ca, c.ca_footer ]

    fname = os.sep.join(pkgnames)
    if fname:
        fname += os.sep
    F_elems[e] = ( fname + e.name + ".cpp", c.ca_list)
    F_elems[id(e)] = (fname + e.name + ".h", ch.ca_list)
    # BUG wrong place
    if F_elems[e][0] == "main.cpp":
        c.ca_imports.add("#include <iterator>");
        c.ca_footer.add("")
        c.ca_footer.add("int main(int argc, char** argv) {")
        c.ca_footer.indent(1)
        c.ca_footer.add("std::vector<std::string> s;")
        c.ca_footer.add("std::copy(argv + 1, argv + argc, std::back_inserter(s));")
        c.ca_footer.add("main::start_program(s);")
        c.ca_footer.add("return 0;")
        c.ca_footer.indent(-1)
        c.ca_footer.add("}")

    hname = "_".join([i.upper() for i in pkgnames])
    if hname:
        hname += "_"
    ch.ca_header.add("#ifndef H_" + hname + str(e.name).upper())
    ch.ca_header.add("#define H_" + hname + str(e.name).upper())
    ch.ca_header.add("")
    ch.ca_footer.add("#endif")

    pkgname = " ".join("namespace " + i + " {" for i in pkgnames)

    l = 16 - len(jhelp.converter_version)
    c.ca_header.add("/***********************************************************************")
    c.ca_header.add(" *                                                                     *")
    c.ca_header.add(" * This file was generated with Coral uml2cpp version " + jhelp.converter_version + " " * l + "*")
    c.ca_header.add(" * DO NOT EDIT!                                                        *")
    c.ca_header.add(" *                                                                     *")
    c.ca_header.add(" ***********************************************************************/")
    if pkgname:
        c.ca_ns_decl.add(pkgname)
        c.ca_ns_decl.add("")
        c.ca_footer.add("")
        c.ca_footer.add("} " * len(pkgnames) )

        ch.ca_ns_decl.add(pkgname)
        ch.ca_ns_decl.add("")
        ch.ca_ns_end.add("")
        ch.ca_ns_end.add(("} " * len(pkgnames))[:-1] )
        
    # CLASS DESCRIPTION
    
    a = ch.ca.add
    i = lambda: ch.ca.indent(1)
    d = lambda: ch.ca.indent(-1)

    s_c = ""
    s_i = ""
    _l = []
    for g in e.generalization:
        _l.append(jhelp.visibility_u2j(g.visibility) + jhelp.sep_name(g.parent, '::'))
        ch.import_these.add(g.parent)
    if _l:
        s_c = " : " + ", ".join(_l)

    typestr = "class"

    jhelp.add_comment(c.ca, e.comment.body)

    # BUG class visibility:  jhelp.visibility_u2j(e.visibility) + 
    a(typestr + " " + e.name + s_c + s_i + " {")

    # ATTRIBUTES
    attr = e.feature.select(lambda x: x.oclIsKindOf(L.Attribute))
    for at in attr:
        create_attribute(ch,
                         at.name,
                         at.type,
                         at.multiplicity,
                         at.ordering,
                         at.visibility,
                         at.comment.body)

    for ae in e.association:
        ae2 = ae.association.connection[0]
        if ae == ae2: ae2 = ae.association.connection[1]
        if not ae2.isNavigable:
            continue
        create_attribute(ch,
                         ae2.name,
                         ae2.participant,
                         ae2.multiplicity,
                         ae2.ordering,
                         ae2.visibility,
                         ae2.comment.body)

    # METHODS
    for m in e.feature.select(lambda x: x.oclIsKindOf(L.Method)):
        create_method(m, c, ch)

    # DEPENDENCIES
    # Also import these classes
    l1,l2 = _get_types(e)
    for i in l1:
        c.import_these.add(i)
        ch.import_these.add(i)
    for i in l2:
        c.import_these.add(i)

    a("")
    a("};")

    def has_special_include_tag(e):
        for i in e.taggedValue:
            if i.name == "include":
                return (True, str(i.dataValue))
        return (False, "")

    def fix(c, declare_forwars = False):
        # IMPORT DECLARATIONS
        s = coral.Set()
        l = []
        for n in c.import_these:
            if n.__getRoot__() == jhelp.cpp:
                continue
            (b, includefile) = has_special_include_tag(n)
            if b:
                includefile = includefile.strip()
                if includefile:
                    l.append(includefile)
            else:
                s.add(n)
        for n in s:
            fwd = ""
            pkglist = coral.Element__getParentsAndMyself__(n, False)
            pkgnames = [ str(i.name) for i in pkglist ]
            pkgname = " ".join("namespace " + i + " {" for i in pkgnames)
            fwd += pkgname
            if fwd:
                fwd += " "
            fwd += "class " + str(n.name) + ";"
            fwd += " }" * (len(pkgnames))
            c.ca_imports.add(fwd)
        if s:
            c.ca_imports.add("")
        names = [ jhelp.sep_name(n, '/') + ".h" for n in s ]
        names.extend(l)
        names.sort()
        for n in names:
            c.ca_imports.add("#include <" + n + ">")
        if names:
            c.ca_imports.add("")

    fix(c)
    # Header does not include itself
    ch.import_these.drop(e)
    fix(ch, True)
    

