"""Some helper stuff for Java handling."""

import coral
import os
import re

converter_version = coral.version

### Some helpers ########################################################

def visibility_u2j(v, whitespace = True):
    """Converts a UML 1.4 visibility string to its code equivalent."""
    if v == "package":
        s = "public"
    else:
        s = str(v)
    if not whitespace:
        return s
    return s + " "

class Creator:
    """Stupid "bunch" class which just can be used to contain anything"""
    def __init__(self):
        pass

def sep_name(e, sep):
    # Shortcircuit cpp::lang::void to become plain void
    if e.__getRoot__() == cpp:
        return str(e.name)
    c = e.taggedValue.select(lambda x: x.name == "name")
    if c:
        return c[0].dataValue
    bindings = e.clientDependency.select(lambda x: x.oclIsKindOf(L.Binding))
    if bindings:
        assert len(bindings) == 1
        tmpl = bindings[0].supplier
        assert len(tmpl) == 1
        tmpl = tmpl[0]
        assert len(tmpl.templateParameter) == len(bindings[0].argument)
        tn = []
        for i in bindings[0].argument:
            s = sep_name(i.modelElement, sep)
            if s:
                tn.append(s)
        template_name = sep_name(tmpl, sep)
        return template_name + "<" + ", ".join(tn) + ">"
    l = coral.Element__getParentsAndMyself__(e)
    s = sep.join( [ str(i.name) for i in l ] )
    return s

def dot_name(e):
    return sep_name(e, '.')

def find(e, s):
    l = s.split(".")
    for i in l:
        e = e.ownedElement.select(lambda x: x.name == i)[0]
    return e

def _remove_empty_prefix_strings(l):
    res = []
    start = True
    for i in l:
        if start and not i:
            continue
        start = False
        res.append(i)
    return res

def add_comment(ca, comment_elems):
    l = []
    for e in comment_elems:
        m = unicode(e).split("\n")
        for s in m:
            l.append(s.rstrip())
        
    if not l:
        return

    l = _remove_empty_prefix_strings(l)
    l.reverse()
    l = _remove_empty_prefix_strings(l)
    l.reverse()
        
    for i, c in enumerate(l):
        s = " * " + c
        s = re.sub("\*\/", "* /", s)
        l[i] = s

    l.insert(0, "/**")
    l.append(" **/")

    for i in l:
        ca.add(i)

def _instantiate_template(template_classifier, classifier_list):
    c = L.Class("dummy-created-by-template-instantiation")
    t = L.TaggedValue(name = "include", dataValue="")
    c.taggedValue.add(t)
    b = L.Binding()
    b.client.add(c)
    b.supplier.add(template_classifier)
    for i in classifier_list:
        a = L.TemplateArgument()
        a.modelElement = i
        b.argument.add(a)
    return c
    

# The +1 blessed greased fake C++ STL environment in UML

L = coral.lang.UML14
_p = coral.loadDocument("coral/addins/cgenUML14/uml2cpp/std.xml")
std = _p.contents().select(lambda x: x.oclIsKindOf(L.Model) \
                           and x.name == "std")[0]
cpp = _p.contents().select(lambda x: x.oclIsKindOf(L.Model) \
                           and x.name == "cpp")[0]
