import coral
import jhelp
import os

L = coral.meta.UML14

from coral.lgen.codeartifact import CodeArtifact

### METHOD ###########################################################

def create_method(method, c):
    c.ca.add("")
    t_static = ""
    if method.specification.ownerScope == "instance":
        pass
    elif method.specification.ownerScope == "classifier":
        t_static = "static "
    else:
        assert 0

    res = method.specification.name
    res += "("
    # Loop through method formal parameters
    params = []
    retparams = []
    for param in method.specification.parameter:
        if param.kind == "return":
            retparams.append(param)
            continue
        params.append(param)

    def f(params, addself = False):
        l = []
        if addself:
            l.append("self")
        for p in params:
            # NO NEED TO IMPORT IN PYTHON
            #c.import_these.add(p.type)
            s = str(p.name)
            if p.defaultValue and p.defaultValue.body:
                s += " = " + p.defaultValue.body
            l.append(s)
        return ", ".join( l )
    
    # BUG loop through keywords args
    # BUG remember to set default values.
    rp = f(retparams)
    #if not rp:
    #    rp += jhelp.dot_name(jhelp.java.lang.void)
    rp += " "
    res = rp + res + f(params, not t_static)

    res += ")"
    res = jhelp.visibility_u2j(method.specification.visibility) + "def" + res
    # COMMENTS
    l = []
    l.extend(method.specification.comment.body)
    l.extend(method.comment.body)
    for param in method.specification.parameter:
        if param.comment:
            s = []
            s.extend(param.comment.body)
            s[0] = "@" + unicode(param.name) + " - " + unicode(s[0])
            l.extend( s )
        
    c.ca.add(res + ":")
    # Method body
    c.ca.indent(1)
    jhelp.add_comment(c.ca, l, True)
    
    b = ""
    if method.body:
        b = unicode(method.body.body).strip()

    if b == "" and method.specification.specification:
        b = unicode(method.specification.specification)

    if b == "":
        c.ca.add("# No method body defined")
        c.ca.add("print \"" + res + "\"")
    else:
        b = b.split("\n")
        for i in b:
            c.ca.add(i)
    c.ca.indent(-1)
    if t_static:
        c.ca.add(method.specification.name + " = staticmethod(" + method.specification.name + ")")

### ATTRIBUTES ###########################################################

def create_attribute(c, name, type, multiplicity, ordering, visibility, comment):
    # No attributes in python?
    return
    l = 0
    u = 1
    if multiplicity:
        for r in multiplicity.range:
            l = min(l, r.lower)
            u = max(u, r.upper)

    if u > 1:
        if ordering == "ordered":
            type = jhelp.java.util.Vector
        else:
            type = jhelp.java.util.TreeSet

    c.import_these.add(type)
    jhelp.add_comment(c.ca, comment)
    c.ca.add(jhelp.visibility_u2j(visibility) + jhelp.dot_name(type) + " " + name + ";")


### CLASS/INTERFACE ######################################################

def create_class(F_elems, e):
    c = jhelp.Creator()
    c.import_these = coral.Set()
    
    pkglist = coral.Element__getParentsAndMyself__(e, False)

    pkgnames = [ str(i.name) for i in pkglist ]

    c.ca_header = CodeArtifact()
    c.ca_imports = CodeArtifact()
    c.ca = CodeArtifact()
    c.ca_footer = CodeArtifact()
    c.ca_list = [ c.ca_header, c.ca_imports, c.ca, c.ca_footer ]
    fname = os.sep.join(pkgnames)
    if fname:
        fname += os.sep
    F_elems[e] = ( fname + e.name + ".py", c.ca_list)
    # BUG wrong place
    if F_elems[e][0] == u"main.py":
        c.ca_footer.add("if __name__ == \"__main__\":")
        c.ca_footer.indent(1)
        c.ca_footer.add("import sys")
        c.ca_footer.add("main.main(sys.argv[1:])")

    # BUG move this to non-existing create_package
    for i in range(len(pkgnames)):
        p = pkglist[i]
        F_elems[p] = ("/".join(pkgnames[0:i+1]) + "/__init__.py", [])

    pkgname = ".".join( pkgnames )

    l = 14 - len(jhelp.converter_version)
    c.ca_header.add("#***********************************************************************")
    c.ca_header.add("#*                                                                     *")
    c.ca_header.add("#* This file was generated with Coral uml2python version " + jhelp.converter_version + " " * l + "*")
    c.ca_header.add("#* DO NOT EDIT!                                                        *")
    c.ca_header.add("#*                                                                     *")
    c.ca_header.add("#***********************************************************************/")
    #if pkgname:
    #    c.ca_header.add("package " + pkgname + ";")
    #    c.ca_header.add("")

    # CLASS DESCRIPTION
    
    a = c.ca.add
    i = lambda: c.ca.indent(1)
    d = lambda: c.ca.indent(-1)

    s_c = ""
    s_i = ""
    _l = []
    for g in e.generalization:
        _l.append(jhelp.dot_name(g.parent) + "." + str(g.parent.name))
        c.import_these.add(g.parent)
    if _l:
        s_c = "(" + ", ".join(_l) + ")"

    typestr = "class"

    a(jhelp.visibility_u2j(e.visibility) + typestr + " " + e.name + s_c + s_i + ":")
    i()

    jhelp.add_comment(c.ca, e.comment.body, True)
    

    did_something_for_class = False
    # ATTRIBUTES
    attr = e.feature.select(lambda x: x.oclIsKindOf(L.Attribute))
    for at in attr:
        #did_something_for_class = True
        create_attribute(c,
                         at.name,
                         at.type,
                         at.multiplicity,
                         at.ordering,
                         at.visibility,
                         at.comment.body)

    for ae in e.association:
        ae2 = ae.association.connection[0]
        if ae == ae2: ae2 = ae.association.connection[1]
        if not ae2.isNavigable:
            continue
        #did_something_for_class = True
        create_attribute(c,
                         ae2.name,
                         ae2.participant,
                         ae2.multiplicity,
                         ae2.ordering,
                         ae2.visibility,
                         ae2.comment.body)

    # METHODS
    for m in e.feature.select(lambda x: x.oclIsKindOf(L.Method)):
        did_something_for_class = True
        create_method(m, c)

    # DEPENDENCIES
    # Also import these classes
    for dep in e.clientDependency:
        for target in dep.supplier:
            c.import_these.add(target)

    if not did_something_for_class:
        a("pass")

    d()
    a("")

    # IMPORT DECLARATIONS
    s = coral.Set()
    for n in c.import_these:
        s.add(n)
    names = [ jhelp.dot_name(n) for n in s ]
    names.sort()
    for n in names:
        # We can't import two "A" classes from different packages
        # using the "import foo.bar.A; import qwe.rty.A;" statements
        # ???
        c.ca_imports.add("import " + n)
    if names:
        c.ca_imports.add("")

