# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

import string
import coral
logger=coral.log.getLogger("autogen")


from coral.modeler import extension
from coral.modeler.diagram import di_support
DI = coral.lang.XMIDI20

empty_list = [ None ]

def get_profile_of_graphelement(ge):
    assert ge.oclIsKindOf(DI.GraphElement)
    p = di_support.find_property_upwards(ge, "x-coral-profile")
    if p:
        return str(p.value)
    a = di_support.findUpwardsModelElement(ge)
    if not a:
        return ""
    L = a.__type__().__language__()
    return L.getName() + str(L.getVersion()) + str(L.getRevision())

class __RendererGenerator:
    def __init__(self):
        self.extensions={}
        # Tuple of (profile, typeinfo) => extension
        self.simpleExtensions = {}
        self.todo_registries = {}
        
    def updateExtensions(self, extRegistry):
        self.todo_registries[extRegistry] = True
        for entry in extRegistry.getRegisteredExtensionsByType(type="DiagramElementExt"):
            ed=extension.fromFullElementName2AbstractElementDefinition(entry[0])
            renderer = entry[1]
            if self.extensions.has_key(ed):
                if renderer not in self.extensions[ed]:
                    self.extensions[ed].append(renderer)
            else:
                self.extensions[ed]=[renderer]
        return

    def __get_simple_from_registries(self, typeinfo, profile):
        t = (profile, typeinfo)
        if self.simpleExtensions.has_key(t):
            return self.simpleExtensions[t]
        for extRegistry in self.todo_registries.keys():
            #print "REG", extRegistry
            for entry in extRegistry.getRegisteredExtensionsByType("SimpleDiagramElementExt", [ profile ]):
                #print "ENTRY", entry[0]
                if entry[0] == typeinfo:
                    self.simpleExtensions[t] = entry[1]
                    return entry[1]
                #if self.simpleExtensions.has_key(entry[0]):
                #    if entry[1] not in self.simpleExtensions[entry[0]]:
                #        self.simpleExtensions[entry[0]].append(entry[1])
                #else:
                #    self.simpleExtensions[entry[0]]=[entry[1]]
            

    def getExtension(self, ge):
        assert ge
        global empty_list
        profile = get_profile_of_graphelement(ge)
        if ge.semanticModel:
            if ge.semanticModel.oclIsKindOf(DI.SimpleSemanticModelElement):
                typeinfo = unicode(ge.semanticModel.typeInfo)
                return self.__get_simple_from_registries(typeinfo, profile)
            elif ge.semanticModel.oclIsKindOf(DI.Uml1SemanticModelBridge):
                if ge.semanticModel.element:
                    elementdefinition = ge.semanticModel.element.__type__()
                    return self.extensions.get(elementdefinition, empty_list)[0]
                    
            else:
                logger.warn("Don't know how to handle" + str(ge) + " " + str(ge.semanticModel))
                assert 0

        return None

    def getFirstSuitableExtension(self, ge):
        """Searches recursevily through the parents of the elementdefinitions to get a suitable renderer.
        For instance a UML14::Model uses a renderer for a UML14::Package"""
        r = self.getExtension(ge)
        if r:
            return r

        lst = []
        if ge.semanticModel and ge.semanticModel.oclIsKindOf(DI.Uml1SemanticModelBridge) and ge.semanticModel.element: 
            lst = ge.semanticModel.element.__type__().getAllSuperclasses()#__superclasses__()

        for sup in lst:
            # return as soon as we find the first suitable extension
            r = self.extensions.get(sup, empty_list)[0]
            if r:
                return r

        return None

RendererGenerator = __RendererGenerator()

