from qt import *
import DependentProjectsDlg as DPDlg

import coral
import os

TM=coral.getTransactionManager()

LEFT=0
RIGHT=1
BOTH=2

class PropertyListItem(QCheckListItem):
    def __init__(self, parent, prj, direction):
        if direction == LEFT:
            QCheckListItem.__init__(self,parent,u"<-- "+unicode(prj.getFilename()),QCheckListItem.CheckBox)
        elif direction == RIGHT:
            QCheckListItem.__init__(self,parent,u"--> "+unicode(prj.getFilename()),QCheckListItem.CheckBox)
        elif direction == BOTH:
            QCheckListItem.__init__(self,parent,u"<-> "+unicode(prj.getFilename()),QCheckListItem.CheckBox)
        else:
            QCheckListItem.__init__(self,parent,u"--- "+unicode(prj.getFilename()),QCheckListItem.CheckBox)

        self.prj=prj

        path=os.path.dirname(str(prj.getFilename()))
        if not path:
            path="."
        if os.access(path, os.W_OK):
            self.setOn(True)
        else:
            self.setOn(False)
            self.setEnabled(False)
        
class DependentProjectsDialog(DPDlg.DependentProjectsDlg):
    def __init__(self, project, parent=None):
        DPDlg.DependentProjectsDlg.__init__(self,parent)
        self.projectToSave = {}
        self.dtable.setTreeStepSize ( 12 )
        self.dtable.setRootIsDecorated(True)
        
        listed=[[project]]
        level=0
        root=PropertyListItem(self.dtable, project, -1)
        listitem={project:root}
        self.projectToSave[root]=project
        self.listitem2children={}

        self.item = None
        
        stop=False
        while not stop:
            listed.append([])
            level+=1
            stop=True
            for prj in listed[level-1]:
                parent=listitem[prj]
                parent.setOpen(True)
                children=[]
                for p,direction in self.adjacent_prjs(prj):
                    stop=False
                    if p in listitem:
                        continue
                    listed[level].append(p)
                    item=PropertyListItem(parent, p, direction)

                    children.append(item)
                    
                    if project.isUserProject() == p.isUserProject():
                        item.setOn(True)
                        item.setEnabled(True)
                        self.projectToSave[item]=p
                    else:
                        item.setOn(False)
                        item.setEnabled(False)

                    listitem[p]=item

                self.listitem2children[parent]=children

    def adjacent_prjs(self,prj):
        # returns project and direction
        referencedProjects  = coral.modeler.project.getReferencedProjects(prj,False)
        referencingProjects = coral.modeler.project.getReferencingProjects(prj,False)

        referencedProjects.pop(prj,None)
        referencingProjects.pop(prj,None)

        # Select arrow type for the project
        prj2arrowtype={}
        for p in referencedProjects:
            if p in referencingProjects:
                prj2arrowtype[p] = BOTH
            else:
                prj2arrowtype[p] = RIGHT

        for p in referencingProjects:
            if p in referencedProjects:
                prj2arrowtype[p] = BOTH
            else:
                prj2arrowtype[p] = LEFT
            

        return prj2arrowtype.items()

    def valueChanged(self, item):
        if not item:
            return
        if not isinstance(item,PropertyListItem):
            return
        
        if item.isOn() and not self.projectToSave.has_key(item):
            self.projectToSave[item] = item.prj
            #for child in self.listitem2children[item]:
            #    child.setOn(True)
            #    self.valueChanged(child)

        elif not item.isOn() and self.projectToSave.has_key(item):
            self.projectToSave.pop(item, None)
            for child in self.listitem2children[item]:
                child.setOn(False)
                self.valueChanged(child)
        
    def getSelected(self):
        return self.projectToSave.values()
         

    def rightClicked(self, item, point, a):
        self.item = item
        menu=QPopupMenu(self,"menu")

        menu.insertItem("Rename project",self.renameProject)
        menu.popup(point,0)
                

    def renameProject(self):
        if not self.item:
            return
        prj = self.item.prj
        import coral.modeler.dlg.ProjectFilename as PF
        dlg=PF.ProjectFilenameDialog()
        dlg.setFilename(str(prj.getFilename()))
        dlg.exec_loop()
        if dlg.result() == dlg.Rejected:
            return
        fn=dlg.selectedFile()
        

        coral.getTransactionManager().beginModifyCmd()
        try:
            prj.setFilename(fn)
        finally:
            coral.getTransactionManager().endModifyCmd()

        self.item.setText(0,unicode(self.item.text()[:4])+fn)
        

