# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.


class EditorSet:
    def __init__(self):
        self.linked=[]
         
    def getEditors(self):
        return self.linked
    
    def link(self, editor):
        if editor not in self.linked:
            self.linked.append(editor)
            editor.editorSet=self
        else:
            print "WARNING, could not link", editor
            print "         to editorSet", self
            print "         because it is already there."
                
    def unlink(self,e):
        assert isinstance(e.editorSet, EditorSet)
        if e in self.linked:
            self.linked.remove(e)
            e.editorSet=None
        else:
            print "WARNING, could not unlink", e
            print "         from editorSet", self
            print "         because it has", e.editorSet
            for i in self.linked:
                print "***", i

    def notifyFocusOn(self,element,exclude=None):
        for e in self.linked:
            if e!=exclude:
                e.focusOn(element)

class Editor:
    """
    Editor is the base class of all editors
    """
    
    def __init__(self, editorSet):
        self.focus=None
        assert editorSet
        assert isinstance(editorSet, EditorSet)
        editorSet.link(self)

    def finalize(self):
        assert isinstance(self.editorSet, EditorSet)
        self.editorSet.unlink(self)
        self.editorSet = None

    def focusOn(self,element):
        """The focus on method tells on which element the editor
        should keep focus on
        """
        self.focus=element

    def apply(self):
        """The apply method applies the changes that the user has made
        to some element, i.e the user has changed the name of a
        Attribute in class, the apply\ method will do the changes on
        the modelElement itself """
        pass

    
    def canEdit(self,e):
        """Return true if the Editor can edit the given element 'e',
        i.e The UMLAssociationEditor canEdit method has the body
        'isinstance(e,Association)'. \ It returns true if the given
        element 'e' is an Association othervise false is returned
        """
        return 0
        
    def notifyFocusOn(self,element):
        """The notifyFocusOn tells the editor on which element and its
        presentation to focus on
        """
        if self.editorSet:
            self.editorSet.notifyFocusOn(element,self)

