# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

import sys
import glob
import os.path
import qt
import traceback
import string
import copy

import coral

DI = coral.lang.XMIDI20


from coral.modeler.window import CoralMainWindow,CoralTabWidget
from coral import modeler
from coral.modeler import iconcache

from coral.modeler.coralmodeler import CoralModeler
from coral.modeler import editor

from coral.modeler import extension

import coral.modeler.project

from coral.log import getLogger
from coral import configuration

_log=getLogger("coral.modeler.modeler")


n = 0
def perpetualEnumerator():
    global n
    n += 1
    return n

def lang_sorter(x, y):
    n = cmp(x.name, y.name)
    if n:
        return n
    n = cmp(x.version, y.version)
    if n:
        return n
    n = cmp(x.revision, y.revision)
    if n:
        return n
    
def chooseLanguage():
    from coral.modeler.dlg.LanguageSelectionDlgEx import LanguageSelectionDlgEx
    gl = [i for i in extension.getLanguages() ]
    gl.sort(lang_sorter)
    lindex=LanguageSelectionDlgEx().main(gl)
    if lindex==-1:
        return None
    return gl[lindex]

def chooseTemplate():
    install_path = coral.installation_path
    user_home = coral.util.user.getUserHome()

    from coral.modeler.dlg.TemplateSelectionDlgEx import TemplateSelectionDlgEx
    
    templatePaths = []
    files2=[]
    templateFiles = {}
    
    templatePaths += [install_path+"/coral/addins/*/mtemplates/*.xml",
                  user_home + "/.coral/coral/addins/*/mtemplates/*.xml"]

    for pattern in templatePaths:
        for f in glob.glob(pattern):
            dirn = os.path.dirname(f)
            files=os.listdir(dirn)
            for foo in files:
                if unicode(foo)[-4:]==".xml":
                    files2.append(foo)
                    
            for fil in files2:
                templateFiles[fil] = dirn+"/"+fil
            files2=[]
            
    tempKey = unicode(TemplateSelectionDlgEx().main(templateFiles))+".xml"
    if not templateFiles.has_key(tempKey):
        return None
    return templateFiles[tempKey]
                
    
        


class ModelerWindow(CoralMainWindow):

    def __showAboutWindow(self, opid):
        from coral.modeler.dlg import AboutDlg
        a = AboutDlg.about(self)
        a.version.setText(unicode(coral.version))
        a.build.setText(unicode(coral.build_date))
        #a.build.hide() # Currently unused
        #a.buildLabel.hide()
        a.show()
        
        
    def __init__(self, show = True):
        CoralModeler.nwindows+=1
        CoralMainWindow.__init__(self)
        
        qp = qt.QPopupMenu(self)
        qp.insertItem("About Coral...", self.__showAboutWindow)
        mb = self.menuBar()
        mb.insertSeparator()
        mb.insertItem("Help", qp)


        # This is the default geometry. Proper widget geometry will be
        # loaded from configuration when the modelerwindow is shown.
        self.resize(600,400)

        self.extensionTools=[]
        self.split = None
        self.split2 = None
        self.split3 = None
        
        coral.getTransactionManager().addBusyObserver(self.__kernelBusy)
        coral.getTransactionManager().addObserver(self.onModelChanged)

        self.focus = None
        self.theEditorSet = editor.EditorSet()
        self.theEditorSet.link(self)
        self.__initExtensions()

        self.contextMenuExtensions={"Edit": [], "Element" : []}
        
        for ext in CoralModeler.extensionRegistry.getRegisteredExtensionsByTypeExt(type="ContextMenuAction"):
            group=unicode(ext.group)
            if not group:
                group="Element"
            if not self.contextMenuExtensions.has_key(group):
                self.contextMenuExtensions[group]=[]

            self.contextMenuExtensions[group].append(CoralModeler.extensionRegistry.getFactory(ext)(self.editorSet))
        self.toolBar.setLabel("Coral")
        for a in ["saveProject","undo","redo"]:
            self.action[a].addTo(self.toolBar)

        for a in ["newProject","newProjectTemplate","coralPathEdit","openProject","cloneWindow","quit"]:
            self.action[a].setEnabled(1)

        self.createEditors()


        self.connect(self.menu["Element"],qt.SIGNAL('aboutToShow()'),self.aboutToShowElementMenu)
        self.connect(self.menu["Edit"],qt.SIGNAL('aboutToShow()'),self.aboutToShowEditMenu)
        
        self.setIcon(iconcache.getIcon("coral"))
        self.main(loop=0)
        if show:
            self.show()
   
    def initActions(self):      
        return [

            modeler.CoralAction(self,"quit",self.quit, "&Quit",
                        qt.QIconSet(iconcache.getIcon("exit")),
                        qt.Qt.CTRL+qt.Qt.Key_Q),

            self.newProject,
            self.newModel,
            modeler.CoralAction(self,"newProjectTemplate",self.newProjectTemplate,"New Project from &Template...",
                    qt.QIconSet(iconcache.getIcon("filenewm")),
                    qt.Qt.CTRL+qt.Qt.Key_T),
            
            modeler.CoralAction(self,"coralPathEdit",self.coralPathEdit,"&Edit coralpath...",
                    qt.QIconSet(iconcache.getIcon("misc")),
                    qt.Qt.CTRL+qt.Qt.Key_E),
            
            modeler.CoralAction(self,"openProject",self.openProject,"&Open Project...",
                    qt.QIconSet(iconcache.getIcon("fileopen")),
                    qt.Qt.CTRL+qt.Qt.Key_O),
            modeler.CoralAction(self,"renameProject",self.renameProject,"&Rename Project...",
                    qt.QIconSet(iconcache.getIcon("misc")),
                    None),
            modeler.CoralAction(self,"saveProject",self.saveProject,"&Save Project",
                    qt.QIconSet(iconcache.getIcon("filesave")),
                    qt.Qt.CTRL+qt.Qt.Key_S),
            modeler.CoralAction(self,"saveProjectAs",self.saveProjectAs,"Save Project &As..",
                    qt.QIconSet(iconcache.getIcon("filesaveas")),
                    0),
            modeler.CoralAction(self,"cloneWindow",self.clone,"&New Window",
                        qt.QIconSet(iconcache.getIcon("clone")),
                        qt.Qt.ALT+qt.Qt.Key_N),
            modeler.CoralAction(self,"saveProjectAndDepending",self.saveProjectAndDepending,"Save Project Dep..",
                    qt.QIconSet(iconcache.getIcon("filesaveas")),
                    0),

            undo,
            redo,
                     
            
            ]

    def initMenu(self):
        return [
            ( "File",( "newProject","newProjectTemplate","coralPathEdit","newModel","openProject","renameProject","saveProject",
                      "saveProjectAs","saveProjectAndDepending","quit" )  ),
            ("Edit",["undo","redo"] ),
            ("Element",[]),
            ( "Tool", []),
            ( "Window",["cloneWindow"])
            ]

    def aboutToShowEditMenu(self):
        menu=self.menu["Edit"]
        menu.clear()
        self.action["undo"].addTo(menu)
        self.action["redo"].addTo(menu)
        if self.focus:
            for o in  self.contextMenuExtensions["Edit"]:
                o.createMenuItem(menu,[self.focus],[])

    def aboutToShowElementMenu(self):
        menu=self.menu["Element"]
        menu.clear()
        if self.focus:
            for o in self.contextMenuExtensions["Element"]:
                o.createMenuItem(menu,[self.focus],[])
                
        
    def newProject(self,language=None):
        """
        CoralAction: newProject
        Group: modeler
        Label: &New Project...
        Icon: filenewm
        Accelerator: CTRL+1
        
        Create a new project
        """

  
        # Create new Project and new top-level element for it.
        # The project has no name, though

        project=None
        if not language:
            language = chooseLanguage()            
        if not language:
            return
        
        project=coral.modeler.project.newProject(language)
        self.notifyFocusOn(project.contents()[0])

        return project
        
    def renameProject(self,project=None):
        if not project:
            if not self.focus:
                return
            project=self.focus.__getRootsProject__()
        
        import coral.modeler.dlg.ProjectFilename as PF
        dlg=PF.ProjectFilenameDialog()
        dlg.setFilename(unicode(project.getFilename()))
        dlg.exec_loop()
        if dlg.result() == dlg.Rejected:
            return
        fn=dlg.selectedFile()
        
        coral.getTransactionManager().beginModifyCmd()
        try:
            project.setFilename(fn)
        finally:
            coral.getTransactionManager().endModifyCmd()
        
    def newModel(self,project=None,language=None):
        """
        CoralAction: newModel
        Label: &New Model...
        Icon: filenew
        Accelerator: CTRL+N
        
        Create a new model
        """
        ## If no project is passed, try to retrieve it from the focus
        if not project:
            project=self.focus.__getRootsProject__()
            ## If no valid element has focus, don't continue
            if not project:
                _log.debug("Cannot insert new model into project that does not exist")
                return
        
        if not language:
            # chose a language        
            language = chooseLanguage()

        new=coral.modeler.project.newModel(project,language)
        self.notifyFocusOn(new)

    def coralPathEdit(self):
        from coral.modeler.dlg.CoralPathDlg import CoralPathDlg
        c=CoralPathDlg().main()
        if not c == -1:
            coral.setCoralPath(c)

    def newProjectTemplate(self):
        a = chooseTemplate()
        if a:
            coral.getTransactionManager().beginModifyCmd()
            try:
                tempPrj = coral.loadDocument(a)
                set = coral.Set()
                for e in tempPrj.contents():
                    set.append(e)
                newSet = coral.ModelCopy(set)                
                prj = coral.Project()
                prj.setFilename("Project%d.xml"%perpetualEnumerator())
                for e in newSet:
                    e.__setProject__(prj)
                tempPrj.invalidate()
            finally:
                coral.getTransactionManager().endModifyCmd()

    def openProject(self,fn=""):
        return coral.modeler.project.openProject(fn)
    
    def saveProject(self):
        if self.focus:
            project = self.focus.__getRootsProject__()
            coral.modeler.project.saveProject(project=project)

    
    def saveProjectAs(self,fileName="",xmiVersion=""):
        if self.focus:
            project = self.focus.__getRootsProject__()
            coral.modeler.project.saveProjectAs(fileName,xmiVersion,project=project)

    def saveProjectAndDepending(self, fileName="", xmiVersion=""):
        if self.focus:
            project = self.focus.__getRootsProject__()
            coral.modeler.project.saveProjectAndDepending(fileName,xmiVersion,project=project)


    def clone(self):
        return ModelerWindow()
        
    def close(self, alsoDelete):    
        CoralMainWindow.finalize(self)
        CoralModeler.nwindows-=1
        coral.getTransactionManager().removeBusyObserver(self.__kernelBusy)
        coral.getTransactionManager().removeObserver(self.onModelChanged)
        #
        # The children's close method modifies the editorSet,
        # so we have to iterate with a copy.
        # And we also have to make sure that @i is still alive.
        # Oh dear this sucks. BUG.
        #
        for i in copy.copy(self.theEditorSet.linked):
            if i != self and i in self.theEditorSet.linked:
                i.close(True)
        # This detects if somebody did not remove themselves.
        assert len(self.theEditorSet.linked) == 1
        assert self.theEditorSet.linked[0] == self
        self.theEditorSet.linked = []
        del self.theEditorSet

        for k,v in self.contextMenuExtensions.iteritems():
            for menuitem in v:
                menuitem.finalize()
        del self.contextMenuExtensions

        if not CoralModeler.nwindows:
            _log.debug("Last window")
            self.__saveMWState()
            ret = qt.QMainWindow.close(self, True)
            CoralModeler.quit()
            return ret
        else:
            # Hm, we call the real method with parameter True
            # to free the resources. I wonder why we @alsoDelete == False
            # at all, it should be True. (??)
            return qt.QMainWindow.close(self, True)


    def __saveMWState( self ):
        def storeWidgetStateViaOperatorLtLt( w, ns ):
            s = qt.QString("")
            ts = qt.QTextOStream(s)
            ts << w
            self.config.add(ns,unicode(s))

        def saveSplitterState( w, ns ):
            self.config.addOption(ns,repr(w.sizes()))

        storeWidgetStateViaOperatorLtLt(self,"GUI.dockstate")
        if self.split:
            saveSplitterState(self.split,"GUI.split")
        if self.split2:
            saveSplitterState(self.split2,"GUI.split2")
        if self.split3:
            saveSplitterState(self.split3,"GUI.split3")

        self.config.addOption(["GUI","geometry"],repr(self.geometry().rect()),
                              False)

        self.config.saveConfig()
        

    def __loadMWState( self ):
        def loadWidgetStateViaOperatorGtGt( w, ns ):
            state = self.config.dGet(ns)
            if state:
                s = qt.QString(state)
                qt.QTextIStream(s) >> w

        def loadSplitterState( w, ns ):
            splitsizes = self.config.dGet(ns)
            if splitsizes:
                w.setSizes(eval(splitsizes))

        if self.config.get("GUI.dont_restore_geometry"):
            return
        
        geom = self.config.get("GUI.geometry")
        if geom:
            apply(self.setGeometry,eval(geom))

        loadWidgetStateViaOperatorGtGt(self,"GUI.dockstate")

        if self.split:
            loadSplitterState(self.split,"GUI.split")
        if self.split2:
            loadSplitterState(self.split2,"GUI.split2")
        if self.split3:
            loadSplitterState(self.split3,"GUI.split3")

    def displayDiagram(self, diagram):
        """Makes the modelerwindow focus on the given Diagram
        by notifying all the editors.
        Returns the DiagramView of the diagram.
        Pumps QT, so the change is visible after returning from
        this function.
        """
        assert diagram.oclIsKindOf(DI.Diagram)
        self.notifyFocusOn(diagram)
        self.processEvents()
        dm = self.getDiagramEditorManager()
        return dm.getEditorForDiagram(diagram)

    def focusOn(self, e):
        """Internal focusOn function used in notification.
        Do not use."""
        self.focus=e
        self.activateActions()
      
    #def getTab(self,x,y):
    #    return self.tabs[(x,y)]
    
    def addToolToTab(self,tab,newTool):
        if isinstance(newTool,qt.QWidget):
            text=newTool.caption()
            if not text:
                text="tool"
            if newTool.icon():
                self.tabs[tab].addTab(newTool,qt.QIconSet(newTool.icon()),text)
            else:
                self.tabs[tab].addTab(newTool,text)
            self.tabs[tab].showPage(newTool)
            newTool.show()
        return newTool
    
    def createEditors(self):
        _log.info("createEditors")
     
        self.tabs={}
        self.ewins=[]
       
        Layout = qt.QGridLayout(self.workspace,1,1,1,1,"Form1Layout")

        self.split=qt.QSplitter(self.workspace)
        Layout.addWidget(self.split,0,0)
        
        self.split3=qt.QSplitter(qt.Qt.Vertical,self.split)

        self.split2=qt.QSplitter(qt.Qt.Vertical,self.split)

        self.tabs[(0,0)] = CoralTabWidget(self.split3)
        self.tabs[(1,0)] = CoralTabWidget(self.split2)
        self.tabs[(0,1)] = CoralTabWidget(self.split3)
        self.tabs[(1,1)]= CoralTabWidget(self.split2)
        

        for t in extension.config.ownedElement.select(lambda c: c.name=="ExtensionTool")[0].ownedElement:
            if not unicode(t.startInTabs):
                continue
            x=1
            y=1
            if "," in unicode(t.startInTabs):
                x=int(string.split(unicode(t.startInTabs),",")[0])
                y=int(string.split(unicode(t.startInTabs),",")[1])
                factory = CoralModeler.extensionRegistry.getFactory(t)
                editor = factory(self.theEditorSet,
                                 self.tabs[(x,y)])
            self.addToolToTab((x,y),editor)

        self.__loadMWState()
        self.split.show()

    def createEditor(self,name):
        editor = \
        CoralModeler.extensionRegistry.getToolByName(name)(self.theEditorSet,
                                                           self.tabs[(1,1)])
        if editor:
            self.addToolToTab((1,1),editor)
            self.theEditorSet.link(editor)

        #print editor, "hit"
            
        return editor
    
    def __kernelBusy(self,isBusy):
        if isBusy:
           qt.QApplication.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        else:
            qt.QApplication.restoreOverrideCursor()
            
    def onModelChanged(self,direc):
        _log.debug("Begin ModelerWindow.onModelChanged")

        tm=coral.getTransactionManager()

        if tm.canUndo():
            self.action["undo"].setEnabled(1)
        else:
            self.action["undo"].setEnabled(0)
        if tm.canRedo():
            self.action["redo"].setEnabled(1)
        else:
            self.action["redo"].setEnabled(0)

        self.activateActions()
        _log.debug("End ModelerWindow.onModelChanged")

    def activateActions(self):
        if self.focus and self.focus.__getRootsProject__():
            self.action["newModel"].setEnabled(1)
            self.action["renameProject"].setEnabled(1)
            self.action["saveProject"].setEnabled(1)
            self.action["saveProjectAs"].setEnabled(1)
            self.action["cloneWindow"].setEnabled(1)
            self.action["saveProjectAndDepending"].setEnabled(1)
        else:
            self.action["newModel"].setEnabled(0)
            self.action["renameProject"].setEnabled(0)                
            self.action["saveProject"].setEnabled(0)
            self.action["saveProjectAs"].setEnabled(0)
            self.action["cloneWindow"].setEnabled(0)
            self.action["saveProjectAndDepending"].setEnabled(0)

    def getCurrentFocus(self):
        return self.focus

    def notifyFocusOn(self, element):
        self.theEditorSet.notifyFocusOn(element)
        
    def __initExtensions(self):
        toolMenu=None
        for x in self.privateMenuBar:
            if x[0]=="&Tool":
                toolMenu=x[1]

        l = []
        for e in CoralModeler.extensionRegistry.getRegisteredExtensionsByType(type="ExtensionTool"):
            if e[1] in self.extensionTools:
                continue
            l.append(e)

        l.sort(lambda a, b: cmp(unicode(a[0]), unicode(b[0])))

        for e in l:
            self.extensionTools.append(e[1])
            tool=ToolLaunch(self,e[1], self.theEditorSet)
            name=unicode(e[0])
            label=name
            name=string.replace(name," ","")
            name = name.encode("UTF-8")
            self.action.__declareAction__(
                modeler.CoralAction(self,"open"+name,tool.launchTool,label,
                                    qt.QIconSet(iconcache.getIcon("tool")),
                                    0))
            
            self.action["open"+name].addTo(toolMenu)                
            self.action["open"+name].setEnabled(1)

    # This method should disappear
    def getOutlineEditor(self):
        # BUG This code is ugly
        from coral.modeler.outline.outlineeditor import OutlineEditor
        for i in self.theEditorSet.linked:
            if isinstance(i, OutlineEditor):
                return i
        return None

    def getDiagramEditorManager(self):
        # BUG This code is ugly
        from coral.modeler.diagram.demanager import DiagramEditorManager
        for i in self.theEditorSet.linked:
            if isinstance(i, DiagramEditorManager):
                return i
        return None


class ToolLaunch:
    def __init__(self, window,tool, editorSet):
        self.window = window
        self.tool = tool
        self.editorSet = editorSet
        assert editorSet
        assert tool
        assert window

    def launchTool(self,tab=(1,1)):
        if issubclass(self.tool, editor.Editor):
            instance = self.tool(self.editorSet, self.window.tabs[tab])
        else:
            instance = self.tool(self.window.tabs[tab])
        self.window.addToolToTab(tab,instance)
        if self.editorSet and isinstance(instance,editor.Editor):
            ##print "link ",instance
            self.editorSet.link(instance)
   

def undo():
    """
    CoralAction: undo
    Label: &Undo
    Icon: undo
    Accelerator: CTRL+Z
    """
    if coral.getTransactionManager().canUndo():
        coral.getTransactionManager().undo()
        
def redo():
    """
    CoralAction: redo
    Label: &Redo
    Icon: redo
    Accelerator: CTRL+SHIFT+Z
    """
    if coral.getTransactionManager().canRedo():
        coral.getTransactionManager().redo()


_a = CoralModeler.getInstance()

def main(filenames):
    global _a
    mw=ModelerWindow()
    
    TM = coral.getTransactionManager()
    # Hide all projects that have been used so far
    if filenames:
        # Wrap them all in one transaction.
        TM.beginModifyCmd()
        try:
            for xmifile in filenames:
                try:
                    p = mw.openProject(xmifile)
                    if not p:
                        continue
                    # BUG update the outliner
                except Exception,e:
                    print "Error loading",xmifile+":",e
                    sys.exit(1)
        finally:                    
            TM.endModifyCmd()
    TM.trimPast()
    
    _a.mainLoop()

    return mw
