# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

import qt
import string

from coral.modeler.unicode import QStringToFileName, QFileDialogToFilename, PyStringToFileName

import coral


n = 0
def perpetualEnumerator():
    global n
    n += 1
    return n

def newProject(language):
   coral.getTransactionManager().beginModifyCmd()
   try:
       project=coral.Project()
       project.setFilename("Project%d.xml"%perpetualEnumerator())
       newModel(project,language)
   finally:
       coral.getTransactionManager().endModifyCmd()
   return project

def newModel(project,language):
    if type(language)==type([]) or type(language)==type((1,2)):
        language=coral.lang.CoralEx10.LanguageDescription(name=language[0],version=language[1],revision=language[2])
    if not language:
        return None

    coral.getTransactionManager().beginModifyCmd()
    try:
        kl=coral.Language.__getLanguage__(str(language.name),language.version,language.revision)
        assert(kl)
        new=kl.getRoot()()
        new.__setProject__(project)
    finally:
            coral.getTransactionManager().endModifyCmd()
    return new

def openProject(fileName=""):
    """Open a project stored in fileName."""
    if not fileName:
        dlg=qt.QFileDialog("", "openProject-filedialog")
        XMIversions=coral.getSupportedIOFormats()
        s = {}
        for format in XMIversions:
            vv = coral.getFileEndingsForIOFormat(format)
            for i in vv:
                s["*." + i] = True
        ss = s.keys()
        ss.sort()
        filters = " ".join(ss)
        dlg.setFilters("Model files (" + filters + ")")
        dlg.setCaption("Load Project")
        dlg.show()
        acc = dlg.exec_loop()
        if not acc == qt.QDialog.Accepted:
            return
        _f = dlg.selectedFile()
        fileName = QFileDialogToFilename(_f)
    else:
        _f = fileName
        fileName = PyStringToFileName(fileName)

    if not fileName:
        return

    project=None
    coral.getTransactionManager().beginModifyCmd()
    try:
        project=coral.loadDocument(fileName)
    finally:                    
        coral.getTransactionManager().endModifyCmd()


    return project

def saveProject(project, xmiversion=None):
    """ Save project"""

    sf = None
    assert project
    
    if project:
        sf = project.getFilename()
        x = project.getIOFormat()
        if x and not xmiversion:
            xmiversion = x

    if not xmiversion:
        xmiversion=coral.getSupportedIOFormats()[-1]

    coral.getTransactionManager().beginModifyCmd()
    try:
        if not sf:
            saveProjectAs(project=project)
        else:
            assert project
            project.setFilename(sf)
            project.setIOFormat(xmiversion)
            coral.saveDocument(project)

    finally:
        coral.getTransactionManager().endModifyCmd()
                
def saveProjectAs(newFileName="",xmiversion=None, project=None):
    assert project

    x = project.getIOFormat()
    if x and not xmiversion:
        xmiversion = x

    ioformats = coral.getSupportedIOFormats()
    if not xmiversion:
        xmiversion = ioformats[-1]

    if not newFileName:
        filtername2coralname = {}
        dlg=qt.QFileDialog("")
        filters=''
        selected = -1
        n = 0
        for s in coral.getSupportedIOFormats():
            if xmiversion == s:
                selected = n
            n += 1
            vv = coral.getFileEndingsForIOFormat(s)
            vv = [ "*." + v for v in vv ]
            name = s + " (" + " ".join(vv) + ")"
            filtername2coralname[name] = s
            if filters:
                filters += ";;" + name
            else:
                filters = name
                
        if selected == -1:
            selected = n - 1

        dlg.setFilters(filters)
        dlg.setSelectedFilter(selected)
        dlg.setMode(qt.QFileDialog.AnyFile)
        dlg.setCaption("Save Project")
        dlg.show()
        dlg.exec_loop()
        if dlg.result() == qt.QDialog.Rejected:
            return
        fn=QFileDialogToFilename(dlg.selectedFile())
        ioformatname = QStringToFileName(dlg.selectedFilter())
        for k,v in filtername2coralname.iteritems():
            if k == ioformatname:
                xmiversion = v
                break
    else:
        fn=PyStringToFileName(newFileName)

    if not fn:
        return

    coral.getTransactionManager().beginModifyCmd()
    try:
        #fixing no extension
        _s = string.split(fn, ".")[-1]
        vv = coral.getFileEndingsForIOFormat(xmiversion)
        if _s not in vv:
            fn += "." + vv[0]

        project.setFilename(fn)
        saveProject(project, xmiversion)
    finally:
        coral.getTransactionManager().endModifyCmd()
    return

def saveProjectAndDepending(newFileName="", xmiversion=None, project=None):
    if not project:
        return
    
    import coral.modeler.dlg.DependentProjects

    dlg=coral.modeler.dlg.DependentProjects.DependentProjectsDialog(project)
    dlg.show()
    dlg.exec_loop()
    if dlg.result() == dlg.Rejected:
        return

    for prj in dlg.getSelected():
        #print prj.getFilename()
        saveProject(prj)
    
def getReferencedProjects(project, transitively=True, dependency=None):
    """get all projects to which project has serializable references
    transitively so that we make sure everything is saved"""

    if not dependency:
        dependency={project:True}
    
    if not project:
        return dependency

    assert project.isTransientProject() == False
    
    prj_all_parts={}
    for root in project.contents():
        for elem in root.__getAllParts__():
            prj_all_parts[elem]=True

    for elem in prj_all_parts:
        for slot in elem.__all_slots__():
            prop=slot.getProperty()
            if prop.isComposition():
                continue
            if prop.getTarget().isDatatypeDefinition():
                continue

            if prop.isAnonymous() or prop.isUnserializable():
                continue

            for e in slot:
                if not prj_all_parts.has_key(e):
                    new_prj = e.__getRootsProject__()
                    if not new_prj:
                        continue
                    if new_prj.isTransientProject():
                        continue
                    if not dependency.has_key(new_prj):
                        dependency[new_prj]=True
                        if transitively:
                            getReferencedProjects(new_prj, transitively, dependency)

    return dependency

def getReferencingProjects(project, transitively=True, dependent=None):
    """get all projects which have serializable references to project
    transitively so that we make sure everything is saved"""
    if not dependent:
        dependent={project:True}

    if not project:
        return dependent
    
    assert project.isTransientProject() == False

    prj_all_parts={}
    for root in project.contents():
        for elem in root.__getAllParts__():
            prj_all_parts[elem]=True

    for elem in prj_all_parts:
        for slot in elem.__all_slots__():
            prop=slot.getProperty()
            if prop.isComposition():
                continue
            if prop.getTarget().isDatatypeDefinition():
                continue

            opposite=prop.getOppositeProperty()
            if opposite and not (opposite.isUnserializable() or opposite.isAnonymous()):
                for e in slot:
                    if not prj_all_parts.has_key(e):
                        new_prj = e.__getRootsProject__()
                        if not new_prj:
                            continue
                    
                        if new_prj.isTransientProject():
                            continue
                        if not dependent.has_key(new_prj):
                            dependent[new_prj]=True
                            if transitively:
                                getReferencingProjects(new_prj, transitively, dependent)

    return dependent
