from cquery import csearch_elem, csearch
from query import search, match_elem
from query import match_query_elem as search_elem
from state import MatchData, QueryState

__all__=["search","search_elem","csearch","csearch_elem","MatchData","QueryState","match_elem"]

__doc__="""
module coral.util.search

Implements sub-graph matching for models

The user has to specify a query, either as a coral.Set of elements or
as a single element of type coral.lang.CQUERY10.Query.

NOTE! All query elements have to be connected so that there is a path
from one element to any other element. Either anonymous or not.

Then, depending on the query type, the user invokes search or csearch
to find mappings to the target model. The mappings are returned in a
list and every mapping is a dictionary where the key is the query
element and the value is the target element.

Method 1: CQUERY10.Query and csearch(query, target, allow_permutations=False).

  Using this method it is possible to store the query as a model,
  including information about ignored properties and if the elements
  are active and/or disjoint.

  The query can be created graphically in coral as long as there are
  graphical representations for all the elements of the query. This
  method allows the user to store all the needed query information
  contained in the Query element.

Method 2: coral.Set, coral.util.search.MatchData and search(match_data, target, query_start).

  If the previous approach is too tedious or if the target model
  contains elements which aren't derived from AnyElement, then
  use the following approach: BUG they all are derived from AnyElement

  Create a coral.Set and add the query elements to the set and invoke
  search, like so:

  coral.util.search(MatchData(query_set), target)
  
  Every type of configuration possible using CQUERY10.Query is also
  possible through the MatchData instance. For example:

  match_data = MatchData(query_set)
  match_data.addIgnored(query_element, "name") #Ignore name 
  match_data.setElemDisjoint(query_element) #Disjoint

  matches = coral.util.search(match_data, target)  

Matching in the opposite direction. If the target element is known but
not a mapping from a query that includes the target element then there
is a function called csearch_elem or search_elem that does the trick.
They are defined in the following way.

  def search_elem(match_data, target_elem)
  def csearch_elem(query, target_elem)

  match_data is of type MatchData, see Method 2 above.
  query is of type CQUERY10.Query

  They both return a list of mappings, since target_elem might be
  included in more than one match.
  
For more information see coral/util/search/README.txt
"""
