# cquery.py Implements matching for queries specified using CQUERY10
#
# Functions:
#             csearch_elem
#             csearch

import coral
from query import *
from state import StatePlan

def OLD_NOT_USED_csearch_elem(query, target_elem):
    """csearch_elem(query, target_elem)
    \tUsed to find a sub-graph in the target_elem model
    \tsuch that the target_elem is included in the match.
    \tThe first parameter (query) has to be of type CQUERY10.Query"""

    if not isinstance(query, coral.Element):
        print "csearch_elem takes an Element as its first parameter not",type(query)
        return []
    if not isinstance(target_elem, coral.Element):
        print "csearch_elem takes an Element as its second parameter"
        return []

    query_set=coral.Set()
    active_qelems=coral.Set()
    for q_elem in query.queryelement:
        query_set.append(q_elem.abstractElement)
        if q_elem.isActive:
            active_qelems.append(q_elem.abstractElement)

    allow_permutations=False
    results=Results(allow_permutations)

    match_data = MatchData(query_set, active_elements=active_qelems)
    for qelem in query.queryelement:
        for meta_prop in qelem.ignore:
            match_data.addIgnored(qelem.abstractElement, meta_prop)

    return match_query_elem(match_data, target_elem)
        

# query is of type CQUERY10::Query and target can be of any
# type inherited from SMD01::AnyElement. The target
# is normally a root element such as a UML14::Model.

def OLD_NOT_USED_csearch(cquery, target, allow_permutations=False):
    """csearch(cquery, target, allow_permutations=False)
    \tUsed to search a target model for occurances of the query pattern
    \tspecified in cquery.
    \tcquery has to be of type CQUERY10::Query and the second parameter
    \tof type CORE01::AnyElement"""

    query_start=cquery.queryelement[0].abstractElement
    query_set=coral.Set()

    same_type_fun = (lambda a, b: a.__type__() == b.__type__())

    alternativeCmpMethods={}
    disjoint_qelems=coral.Set()
    for qelem in cquery.queryelement:
        if not qelem.abstractElement:
            assert 0, unicode(qelem)+" does not have an abstractElement"
        query_set.append(qelem.abstractElement)
        if qelem.isDisjoint:
            disjoint_qelems.append(qelem.abstractElement)

        # Check which elem matching function should be used here
        # uncomment and add comparisonmethod to QueryElement if someone will need it. 
        #if qelem.comparisonmethod == "isTypeOf":
        #    alternativeCmpMethods[qelem.abstractElement]=same_type_fun


    ### Find the active element with the highest degree
    active_elements=[]
    for qelem in cquery.queryelement:
        if qelem.isActive:
            active_elements.append(qelem.abstractElement)
            #query_start=qelem.abstractElement
            #break

    max_degree = 0
    for qs_elem in active_elements:
        this_elem_degree=0
        for slot in qs_elem.__all_slots__():
            if slot.getProperty().getTarget().isDatatypeDefinition():
                this_elem_degree += len(slot)
            else:
                this_elem_degree += len(filter(lambda x: x in query_set,slot))

        if this_elem_degree > max_degree:
            query_start = qs_elem
            max_degree = this_elem_degree

    ### Find end

    match_data = MatchData(query_set,
                           allow_permutations = allow_permutations,
                           elem_cmp_fun = alternativeCmpMethods,
                           disjoint_elements = disjoint_qelems)

    for qelem in cquery.queryelement:
        for meta_prop in qelem.ignore:
            match_data.addIgnored(qelem.abstractElement, meta_prop)

    #print "Query start",query_start
    return search(match_data, target, query_start)

##
## New implementation of search, match and state which enables searchplanning
##

def csearch_plan(cquery, target, allow_permutations=False):
    """csearch_plan(cquery, target, allow_permutations=False)
    \tUsed to search a target model for occurances of the query pattern
    \tspecified in cquery. This version uses searchplans and is capable
    \tof negative matching.
    \tcquery has to be of type CQUERY10::Query and the second parameter
    \tof type CORE01::AnyElement"""
    #query_start=cquery.queryelement[0].abstractElement
    query_set=coral.Set()

    alternativeCmpMethods={}
    disjoint_qelems=coral.Set()
    negative_qelems=coral.Set()
    active_elements=coral.Set()
    for qelem in cquery.queryelement:
        if not qelem.abstractElement:
            assert 0, unicode(qelem)+" does not have an abstractElement"
        if qelem.isNegative:
            negative_qelems.append(qelem.abstractElement)
            continue
        
        if qelem.isActive:
            active_elements.append(qelem.abstractElement)

        query_set.append(qelem.abstractElement)
        if qelem.isDisjoint:
            disjoint_qelems.append(qelem.abstractElement)

        # Check which elem matching function should be used here
        # uncomment and add comparisonmethod to QueryElement if someone will need it. 
        #if qelem.comparisonmethod == "isTypeOf":
        #    alternativeCmpMethods[qelem.abstractElement]=same_type_fun

    match_data = MatchData(query_set,
                           allow_permutations = allow_permutations,
                           elem_cmp_fun = alternativeCmpMethods,
                           disjoint_elements = disjoint_qelems,
                           negative_elements = negative_qelems,
                           active_elements = active_elements)

    for qelem in cquery.queryelement:
        for meta_prop in qelem.ignore:
            match_data.addIgnored(qelem.abstractElement, meta_prop)


    # initialize the result collection
    results = Results(allow_permutations)

    # match every possible target element to the query start element.

    state = StatePlan(match_data)

    query_start = state.searchPlan[0][0]

    target_starts = target.__getAllParts__().select(lambda p: match_data.cmp_elem_fun(query_start, p))

    for target_start in target_starts:
        state1 = state.clone()
        
        if state1.isFeasiblePair(query_start, target_start):
            state1.addPair(query_start, target_start)
            match_plan(query_start,target_start, state1)
            
    for res in state.getResults():
        results.append(res)

    return results.getList()

csearch = csearch_plan

def csearch_plan_elem(query, target_elem):
    """csearch_plan_elem(query, target_elem)
    \tUsed to find a sub-graph in the target_elem model
    \tsuch that the target_elem is included in the match.
    \tThe first parameter (query) has to be of type CQUERY10.Query"""

    if not isinstance(query, coral.Element):
        print "csearch_elem takes an Element as its first parameter not",type(query)
        return []
    if not isinstance(target_elem, coral.Element):
        print "csearch_elem takes an Element as its second parameter"
        return []

    query_set=coral.Set()
    active_qelems=coral.Set()
    for q_elem in query.queryelement:
        query_set.append(q_elem.abstractElement)
        if q_elem.isActive:
            active_qelems.append(q_elem.abstractElement)

    allow_permutations=False
    results=Results(allow_permutations)

    match_data = MatchData(query_set, active_elements=active_qelems)
    for qelem in query.queryelement:
        for meta_prop in qelem.ignore:
            match_data.addIgnored(qelem.abstractElement, meta_prop)

    #return match_query_elem(match_data, target_elem)

    # from match_query_elem

    results=Results(match_data.allow_permutations)

    active_qelems=match_data.active_elements
    if not active_qelems:
        active_qelems=match_data.query_set

    for q_elem in active_qelems:
        state = StatePlan(match_data)
        state.searchPlan.newPlan(q_elem)
        if state.isFeasiblePair(q_elem, target_elem):
            state.addPair(q_elem, target_elem)
            match_plan(q_elem, target_elem, state)
            rs=state.getResults()
            if rs:
                for result in rs:
                    results.append(result)

    return results.getList()

csearch_elem = csearch_plan_elem




# #
# # Old implementation that modified the query graph before searching.
# #


# def csearch_elemx(query, elem):
#     if not isinstance(query, coral.Element):
#         print "csearch_elem takes an Element as its first parameter not",type(query)
#         return []
#     if not isinstance(elem, coral.Element):
#         print "csearch_elem takes an Element as its second parameter"
#         return []

#     # Method 1: Prune query graph and use search as is
#     query_set=get_pruned_query_set(query)
#     return  match_query_elem(query_set, elem)

#     # Method 2: Populate the skip list used in state with elements from ignored slots
#     #           Don't use search but implement the same functionallity here..
#     # Not implemented
    
    

# def csearchx(query, target, allow_permutations=False):
#     # Uses search_isKindOf
#     if not isinstance(query, coral.Element):
#         print "csearch takes an Element as its first parameter not",type(query)
#         return []
#     if not isinstance(target, coral.Element):
#         print "csearch takes an Element as its second parameter"
#         return []

#     # Method 1: Prune query graph and use search as is
#     query_set=get_pruned_query_set(query)
    
#     first_active_elem=None
#     for qelem in query.queryelement:
#         if qelem.isActive:
#             first_active_elem=qelem.abstractElement
#             break

#     return search_isKindOf(query_set, target, first_active_elem, allow_permutations)



# def get_pruned_query_set(query):#Tested
#     # TODO: check that pruning doesn't split the query graph
#     #       Do this by checking if the slot contains query elements for instance.

#     # TODO: Maybe model copy the query since pruning affects the original query elements

#     query_set=coral.Set()
#     # TODO: What if we want containedElements?
#     for query_element in query.queryelement:
#         pruned_element=prune_ignore_properties(query_element)
#         query_set.append(pruned_element)
        
#     return query_set

# def prune_ignore_properties(query_element):#Tested
#     element=query_element.abstractElement
#     for prop in query_element.ignore:
#         slot=element.__getSlotByName__(unicode(prop.name))
#         slot_elements=slot[:]
#         for slot_element in slot_elements:
#             slot.remove(slot_element)
#         assert len(slot)==0
#     return element
