

class TraversalNode:
    def __init__(self, element, propName=None, parent=None):
        assert element
        self.parent = parent
        self.element = element
        self.propName = propName
        self.children=[]
        if parent:
            parent.children.append(self)

    def breadthFirst(self, thePlan):
        if not thePlan:
            thePlan.append( (self.element, None, None) )

        for child in self.children:
            thePlan.append( (child.element, self.element, child.propName) )

        for child in self.children:
            child.breadthFirst(thePlan)
            

    def depthFirst(self, thePlan):
        if not thePlan:
            thePlan.append( (self.element, None, None) )

        for child in self.children:
            thePlan.append( (child.element, self.element, child.propName) )
            child.depthFirst(thePlan)

class SearchPlan:
    def __init__(self, match_data):
        self.match_data = match_data
        self.thePlan = self.initSearchPlan() 
        self.theNegativePlan = self.initNegativePlan()

    def __len__(self):
        return len(self.thePlan)

    def at(self, i):
        return self.thePlan[i]

    def __getitem__(self, i):
        return self.at(i)

    def useNegative(self, flag=True):
        if flag:
            self.thePlan = self.thePlan + self.theNegativePlan
        else:
            #self.thePlan = self.thePlan[:-(len(self.theNegativePlan))]
            self.thePlan = self.thePlan[:(len(self.match_data.query_set))]
        
    def adjacent(self, element):
        prop2adjacent={}

        for prop in self.match_data.elem_property_lists[element][0].iterkeys():
            propName = prop.getName()
            prop2adjacent[propName] = []
            for elem in element.__getSlotByName__(propName):
                if elem in self.match_data.query_set:
                    prop2adjacent[propName].append(elem)
            
        # return {"prop":[elems]}
        return prop2adjacent

    def newPlan(self, start_elem):
        self.thePlan = self.initSearchPlan(start_elem)

    def initSearchPlan(self, fromElem=None):

        # Build a tree with a node for each element. All nodes except
        # the root knows which propName to use to get from the parent
        # node. The searchplan is made differently depending on how
        # the tree is traversed.

        query_set = self.match_data.query_set

        start_elem = query_set[0]
        if self.match_data.active_elements:
            start_elem = self.match_data.active_elements[0]

        if fromElem:
            start_elem = fromElem
            
        root = TraversalNode(start_elem)

        planned_elements = {}
        planned_elements[start_elem]=root

        while len(planned_elements) < len(query_set):

            for elem in planned_elements.keys(): # Find adjacent

                for propName, adj in self.adjacent(elem).items():

                    for adj_elem in adj:

                        if adj_elem not in planned_elements: # New element

                            planned_elements[adj_elem]=TraversalNode(adj_elem,propName,planned_elements[elem])

        thePlan=[]

        root.depthFirst(thePlan)

        #root.breadthFirst(thePlan)

        # thePlan=[(new,matched,connecting_prop),...]
        #print thePlan
        return thePlan

    def initNegativePlan(self):
        # 
        # Find a traversal path through the negative parts
        #

        unplanned_elements=[]
                
        new_elements=[]

        thePlan = []

        for qelem in self.match_data.query_set:
            for nelem in self.match_data.negative_elements: # connect negative to query
                propName = self.fromTo(qelem, nelem)
                if propName:
                    thePlan.append( (nelem, qelem, propName) )
                    new_elements.append(nelem)

        for nelem in self.match_data.negative_elements:
            if not nelem in new_elements:
                unplanned_elements.append(nelem)
                    
        while unplanned_elements:
            for planned_nelem in new_elements:
                for nelem in unplanned_elements: # connect negative to negative
                    propName = self.fromTo(planned_nelem, nelem)
                    if propName:
                        thePlan.append( (nelem, planned_nelem, propName) )
                        new_elements.append(nelem)

        #print thePlan
        return thePlan

    def fromTo(self, one, two):
        for slot in one.__all_slots__():
            if two in slot:
                return slot.getProperty().getName()
        return None

    
#         planned_elements = {}

#         for qelem in self.match_data.query_set:
#             planned_elements[qelem] = TraversalNode(qelem)
            
#         negative_elements = self.match_data.negative_elements
#         new_elements = {}
#         # while len(new_elements) < len(negative_elements):

#         for elem in planned_elements.keys():
#             for prop, adj in self.adjacent(elem):
#                 for new_element in new_elements:
#                     if new_element in adj:
#                         thePlan.append( (new_element, elem, prop) )
# ##############################################

