import coral
from state import QueryState, MatchData, StatePlan
from util import Results


def match_query_elem(match_data, target_elem):
    """match_query_elem(match_data, target_elem)
    \tThis function is used to find match(es) which include target_elem"""
    # See if the target element matches any of the query elements

    results=Results(match_data.allow_permutations)

    active_qelems=match_data.active_elements
    if not active_qelems:
        active_qelems=match_data.query_set

#    for q_elem in match_data.query_set:
    for q_elem in active_qelems:
        state=match_data.state_T(match_data)
        if state.isFeasiblePair(q_elem, target_elem):
            state.addPair(q_elem, target_elem)
            match(q_elem, target_elem, state)
            rs=state.getResults()
            if rs:
                for result in rs:
                    results.append(result)

    return results.getList()


def match_elem(query_elem, target_elem, match_data):
    """match_elem(query_elem, target_elem, match_data) returns a list
    \tof matches where query_elem maps to target_elem."""
    # NOTE: match_elem used to match elements which were part of a query element but not in the query_set
    # NOTE: however, this is not the case anymore so make sure to include all elements in query_set

    if not match_data.cmp_elem_fun(query_elem, target_elem):
        return False

    state=match_data.state_T(match_data)

    match(query_elem, target_elem, state)

    return state.getResults()

def search(match_data, target_model, query_start=None):
    """search(match_data, target_model, query_start=None)
    \tfinds all possible mappings from the query in match_data
    \tto the elements in the target_model."""
    assert isinstance(match_data, MatchData)
    # NOTE: search used to match elements which were part of a query element but not in the query_set
    # NOTE: however, this is not the case anymore so make sure to include all elements in query_set

    if not issubclass(match_data.state_T,QueryState):
        assert 0
        
    query_set=match_data.query_set
    
    if not query_set or not target_model: return []

    # if the caller didn't specify the start point then choose the one with the highest degree
    if not query_start:
        query_start = query_set[0]
        max_degree = 0
        for qs_elem in query_set:
            this_elem_degree=0
            for slot in qs_elem.__all_slots__():
                if slot.getProperty().getTarget().isDatatypeDefinition():
                    this_elem_degree += len(slot)
                else:
                    this_elem_degree += len(filter(lambda x: x in query_set,slot))
            if this_elem_degree > max_degree:
                query_start = qs_elem
                max_degree = this_elem_degree
    else:
        assert query_start in query_set

    # find all elements in the target that might match to the query element.
    target_starts=target_model.__getAllParts__().select(lambda p: match_data.cmp_elem_fun( query_start, p))

    # initialize the result collection
    results = Results(match_data.allow_permutations)

    # match every possible target element to the query start element.

    state=match_data.state_T(match_data)

    for target_start in target_starts:
        #print "Try next start point",target_start
        state.reset()
        
        if state.isFeasiblePair(query_start, target_start):
            state.addPair(query_start, target_start)
            match(query_start,target_start, state)
            
    for res in state.getResults():
        results.append(res)

    return results.getList()

def compatible_slots(q_elem, t_elem, state):

    for prop, num in state.element_properties(q_elem).iteritems():
        t_slot = t_elem.__getSlotByName__(prop.getName())
        if len(t_slot) < num:
            return False

    return True

def match(q_elem, t_elem, state, match_primitive = True):
    if isinstance(q_elem,coral.ElementPrimitive):
        assert 0

    # Start by matching only the slots which contain ElementPrimitives.
    if match_primitive:
        for primitive_prop in state.primitive_properties(q_elem):
            propName=primitive_prop.getName()
            q_slot=q_elem.__getSlotByName__(propName)
            t_slot=t_elem.__getSlotByName__(propName)
            
            for q_item,t_item in map(None,q_slot,t_slot):
                if not q_item == t_item:
                    if (len(str(q_item)) and str(q_item)[0] != "*"): # Remove this line when wildcard is not used anymore
                        return False

    if match_primitive and not compatible_slots(q_elem, t_elem, state):
        return False

    # Next up are the slots which contain Elements.
    for prop in state.element_properties(q_elem).iterkeys():

        propName=prop.getName()
        q_slot=q_elem.__getSlotByName__(propName)
        t_slot=t_elem.__getSlotByName__(propName)

        for q_item in q_slot:

            if not q_item in state.match_data.query_set:
                continue
            
            if state.q2t.has_key(q_item): # Has it been matched already?
                if state.q2t[q_item] in t_slot: # Does the t_elem have the same edge
                    continue
                else:
                    return False
                
            for t_item in t_slot:
                if t_item in state.q2t.values(): # This belongs to someone already
                    continue

                if state.isFeasiblePair(q_item,t_item): # Can we add it to the set?
                    state2=state.clone()
                    state2.push(q_elem, t_elem)
                    state2.addPair(q_item,t_item) 
                        
                    if match(q_item,t_item,state2):
                        return True

            return False

                    # every edge has been matched previously
    if state.stack_size()==0:      # we are in the element that we started from
        if state.isGoal():         # did we find mappings for all query elements?
            return state.visitor() # return false to get more matches
        else:
            print "query.py: Your query is not fully traversable!"
            return True
    else:
        a,b=state.pop()
        return match(a,b,state,False)





########## THE NEW MATCH #################################
# Requires a state that is able to provide a search plan #
# such a state "StatePlan" can be found in state.py      #
##########################################################

def match_plan(q_elem, t_elem, state):
    assert isinstance(state, StatePlan)
    assert not isinstance(q_elem,coral.ElementPrimitive)

    # Check constraints that all previously matched elements which are
    # adjacent to the pattern node are also adjacent to the target
    # node. and primitive stuff.
    if not state.check_constraints(q_elem, t_elem):
        return False

    #### Find next pair and call match ####

    new_nextTo_propName = state.nextQueryNode()# From the searchplan

    if new_nextTo_propName:
        q_item, nextTo, propName = new_nextTo_propName

        t_slot = state.q2t[nextTo].__getSlotByName__(propName)

        for t_item in t_slot:
            if t_item in state.q2t.values(): # This belongs to someone already
                continue

            if state.isFeasiblePair(q_item,t_item): # Can we add it to the set?
                state1=state.clone()

                state1.addPair(q_item,t_item) 
                        
                if match_plan(q_item,t_item,state1):
                    return True

        return False
    
    if state.start_matching_negative():# Returns true if it wasn't already matching negative.
        before = len(state.q2t)

        if state.match_data.negative_elements:
            state1 = state.clone()
            if match_plan(q_elem, t_elem, state1):
                state.stop_matching_negative()
                return False

        state.stop_matching_negative()
        assert len(state.q2t) == before
        
        #print "Weird",state.q2t
        
        # every edge has been matched previously
        if state.isGoal():         # did we find mappings for all query elements?
            return state.visitor() # return false to get more matches
        else:
            print "query.py: Your query is not fully traversable!"
            return True

    else:
        return True # A negative match


#     # every edge has been matched previously
#     if state.isGoal():         # did we find mappings for all query elements?
#         return state.visitor() # return false to get more matches
#     else:
#         print "query.py: Your query is not fully traversable!"
#         return True

#     assert 0,"Somehow match reached the end"


