import coral
import coral.util.search as Q
import coral.util.search.util as U

#from coral.modeler.diagram.de import suitesDiagram
from coral.modeler.diagram.degen import DiagramElementGenerator as DEGEN
from code import InteractiveInterpreter
import string
import sys

XMIDI=coral.lang.XMIDI20

TM = coral.getTransactionManager()

class II(InteractiveInterpreter):

    def __init__(self, locals = {}):
        self.fail = False
        self.message = []
        InteractiveInterpreter.__init__(self, locals)
        
    def write(self, data):
        self.message.append(data)
        self.fail = True
        
def swap_key_value(dict):
    ret={}
    for k,v in dict.items():
        ret[v]=k
    return ret

def transform(model, lhs, rhs): #Backwards compatibility
    cquery_transform(model, lhs, rhs)

def getAveragePositionInDiagram(match, diag):
    aX,aY = 0,0
    number=0
    for orig, elem in match.items():
        graph_elements = elem.xmidi_get_graphelements()
        if graph_elements:
            if graph_elements[0].oclIsKindOf(XMIDI.GraphEdge):
                continue

        for gnode in graph_elements:
            if gnode.__isPart__(diag):
                #In case there is a graphnode with ForegroundColor set then use that as the point
                if orig.xmidi_get_graphelements().exists(lambda ge: ge.property.exists(lambda p: str(p.key) == "ForegroundColor")):
                    return gnode.position.x, gnode.position.y

                #Otherwise calculate the average
                aX += gnode.position.x
                aY += gnode.position.y
                number += 1
                break

    if number > 0:
        aX = aX / number
        aY = aY / number
        
    return aX,aY

def addDiagramInfo(model, copy2orig, match):

    # Assumes that the RHS has 0..1 GraphNodes per Element. Creates
    # graphnodes for the added elements. Puts them in every diagram
    # that has graphnodes for the matched elements. Sets position
    # according to the matched elements and the position of the new
    # element in the RHS.
    
    for diag in model.xmidi_get_diagrams():

        rx,ry = getAveragePositionInDiagram(match, diag)

        TM.beginModifyCmd()
        try:  
            edges={}
            for new_elem, orig_elem in copy2orig.items():
                ext = DEGEN.getExtension(new_elem.__type__())
                if not ext:
                    continue
            
                de = ext.createDiagramElement(new_elem, xparent=diag)
                diag.contained.append(de)
                if de.oclIsKindOf(XMIDI.GraphEdge):
                    edges[new_elem]=ext
                else:
                    gelems=orig_elem.xmidi_get_graphelements()
                    if gelems:
                        old_de=gelems[0]
                        px = old_de.position.x + rx
                        py = old_de.position.y + ry
                        de.position = XMIDI.Point(x=px,y=py)

            for elem,ext in edges.items():
                ext.attachGraphConnector(elem)

        finally:
            TM.endModifyCmd()  

def cquery_transform(model, lhs, rhs, query=None, prescript=None, postscript=None):
    
#    print "transform begin"
    rhs2lhs_map=intersection(rhs, lhs) # RHS /\ LHS
    rhs_to_add=minus(rhs, lhs) # RHS \ LHS

    # The next two might not have to be separate
    # 
    lhs_to_del=minus(lhs, rhs) # LHS \ RHS
    element_propname_items_removed = slot_diff(swap_key_value(rhs2lhs_map)) # Check for slot differences
    element_propname_items_added = slot_diff(rhs2lhs_map) # Check for slot differences

    if not query:
        #print lhs_to_del
        matches=Q.search(Q.MatchData(U.get_all_parts(lhs),
                                     #disjoint_elements=lhs_to_del
                                     ),
                         model)
    else:
        matches=Q.csearch(query, model)

    # Compile the script if such exists
    post_executable_code=None
    pre_executable_code=None
    if prescript:
        pre_executable_code = compile(str(prescript), "script", "exec")
    if postscript:
        post_executable_code = compile(str(postscript), "script", "exec")

    # print "Found",len(matches),"matches in",model
    for match in matches:
        rewrite(model, match, pre_executable_code, post_executable_code, rhs2lhs_map, element_propname_items_removed, element_propname_items_added, rhs_to_add, lhs_to_del)

def rewrite(model, match, pre_executable_code, post_executable_code, rhs2lhs_map, element_propname_items_removed, element_propname_items_added, rhs_to_add, lhs_to_del):
    script_locals={}
    # Execute the post script
    if pre_executable_code:
        # Create two dictionaries LHS and RHS which have the matched elements and the new RHS elements
        LHS={}


        for common_elem in rhs2lhs_map.values():
            if match.has_key(common_elem):
                LHS[unicode(common_elem.name)] = match[common_elem]

        for del_elem in lhs_to_del:
            LHS[unicode(del_elem.name)] = match[del_elem]
                
        script_locals["LHS"]=LHS
            
        ii=II(script_locals)
        #mb = MessageBuffer()
        #sys.stdout = mb
        try:
            ii.runcode(pre_executable_code)
        finally:
            sys.stdout = sys.__stdout__
           
        if ii.fail:
            print "Script failed with", string.join(ii.message)

    ### REMOVE Associations from slots that have changed
    for k,items in element_propname_items_removed.items():
        (lhs_elem,prop_name)=k
        if not match.has_key(lhs_elem):
            continue
        real_elem=match[lhs_elem]
        slot=real_elem.__getSlotByName__(prop_name)
        for remove_item in items:
            #print "item slot",remove_item, slot.getProperty().getName()
                
            slot.drop(match[remove_item])

    ### ADD Associations in slots that have changed
    for k,items in element_propname_items_added.items():
        (rhs_elem, prop_name)=k
        #print "Slot grew ",prop_name
        # Check if items are other intersection elements otherwise continue
        lhs_elem = rhs2lhs_map[rhs_elem]
        if lhs_elem not in match:
            #print "Not found ",lhs_elem,"in match"
            continue
            
        for item in items:
            if not item in rhs2lhs_map:
                #print " external element discarding ", item
                continue
                
            #print " common element, adding", item," to slot ",prop_name,"in",lhs_elem
            slot = match[lhs_elem].__getSlotByName__(prop_name)
            slot.insert(match[rhs2lhs_map[item]])

    ### DELETE Elements that should be deleted, 
    deleted_elements=coral.Set()

    for del_item in lhs_to_del:
        # TODO: remove associations that have changed
        # print "Removing",match[del_item]
        if match[del_item] in deleted_elements:
            print "This element was already removed\n **",match[del_item],"**"
        else:
            deleted_elements.append(match[del_item])
            match[del_item].__parentSlot__().resetSlot()
                
    # 3. Glue copy of R\L (to match[L /\ R])

    ### ADD Associations to slots in elements that should be connected
    copy2orig={}
    rhs_to_add_copy=coral.ModelCopy(rhs_to_add,0,{},copy2orig)
    
    for new_elem in rhs_to_add_copy:
        original=copy2orig[new_elem]
        if original and original.__parent__() in rhs2lhs_map:
            p=match[rhs2lhs_map[original.__parent__()]]
            new_elem.__parentSlot__().insert(p)
        else:
            new_elem.__parentSlot__().insert(model) # Move element from RHS to model

        # print "Added ",new_elem
        re_connect(new_elem, match, rhs2lhs_map)

    addDiagramInfo(model, copy2orig, match)


    # Execute the post script
    if post_executable_code:
        # Create two dictionaries LHS and RHS which have the matched elements and the new RHS elements
        RHS={}

        for new_elem in rhs_to_add_copy:
            RHS[unicode(new_elem.name)] = new_elem

        for common_elem in rhs2lhs_map:
            RHS[unicode(common_elem.name)] = common_elem
                
        script_locals["RHS"]=RHS
        script_locals.pop("LHS")
            
        ii=II(script_locals)
        #mb = MessageBuffer()
        #sys.stdout = mb
        try:
            ii.runcode(post_executable_code)
        finally:
            sys.stdout = sys.__stdout__
           
        if ii.fail:
            print "Script failed with", string.join(ii.message)
       
#         #print "transform end"



def slot_diff(lhs2rhs_map):
    #Check all association property slots if they have changed.
    element_propname_items={}
    for l_elem, r_elem in lhs2rhs_map.items():
        for prop in l_elem.__properties__():
            if prop.isAnonymous():
                continue
            if not prop.isAssociation():
                continue
            l_slot=l_elem.__getSlotByName__(prop.getName())
            r_slot=r_elem.__getSlotByName__(prop.getName())
            diff=slot_diff2(l_slot, r_slot, lhs2rhs_map)
            if diff:
                #print diff
                element_propname_items[(l_elem,prop.getName())]=diff
                
    return element_propname_items

def slot_diff2(slot1, slot2, l2rmap):
    removed_items=[]
    for l_item in slot1:
        if not l_item in l2rmap: # Like 1.A[C] on paper
            continue
        if not l2rmap[l_item] in slot2: # l_item does not exist in rhs:s slot
            removed_items.append(l_item)

    return removed_items
        
def re_connect(new_elem, match, rhs2lhs_map):
    # This needs to be done recursively
    if isinstance(new_elem, coral.ElementPrimitive):
        return

    for prop in new_elem.__properties__():
        if prop.isAnonymous():
            continue
        slot=new_elem.__getSlotByName__(prop.getName())
        if prop.isAssociation():
            for item in slot[:]:
                if item in rhs2lhs_map:
                    slot.remove(item)
                    slot.insert(match[rhs2lhs_map[item]])
        else:
            for item in slot:
                re_connect(item, match, rhs2lhs_map)


def intersection(lhs, rhs, common=None):
    # Elements that exist in both
    if common==None:
        common={}
        
    for lhs_elem in lhs:
        if isinstance(lhs_elem, coral.ElementPrimitive):
            continue
        if not hasattr(lhs_elem, "name"):
            continue
        exists=False
        for rhs_elem in rhs:

            if lhs_elem.name == rhs_elem.name:
                intersect_contained(lhs_elem, rhs_elem, common)
                exists=True
                break

        if exists:
            common[lhs_elem]=rhs_elem
            
    return common

def intersect_contained(lhs_elem, rhs_elem, common):
    for prop in lhs_elem.__properties__():
        if prop.isAssociation():
            continue
        lhs=lhs_elem.__getSlotByName__(prop.getName())
        rhs=rhs_elem.__getSlotByName__(prop.getName())
        intersection(lhs,rhs,common)    

def minus(lhs, rhs):
    # Elements that exist in LHS but not in RHS
    diff=coral.Set()
    
    for lhs_elem in lhs:
        exists=False
        if not hasattr(lhs_elem,"name"):
            continue
        for rhs_elem in rhs:

            if lhs_elem.__type__() == rhs_elem.__type__():
                if lhs_elem.name != rhs_elem.name:
                    #print "Different types",lhs_elem,rhs_elem
                    continue
                exists=True
                diff.unionWith(minus_contained(lhs_elem, rhs_elem))
                break

        if not exists:
            diff.append(lhs_elem)
            
    return diff

def minus_contained(lhs_elem, rhs_elem):
    diff=coral.Set()
    for prop in lhs_elem.__properties__():
        if prop.getTarget().isDatatypeDefinition():
            continue
        if prop.isAssociation():
            continue
        lhs=lhs_elem.__getSlotByName__(prop.getName())
        rhs=rhs_elem.__getSlotByName__(prop.getName())
        diff.unionWith(minus(lhs,rhs))

    return diff

# Functions suitable with the DPOTRANS10 language.
# Both LHS and RHS are represented by queries.


import coral.modeler.modeler
import coral.addins.MATCH10.evaluate
    
def DPOminus(left_query, right_query):
    """Determine mappings based on the name of the QueryElement and
    return the queryelements in parameter 1 that do not exist in
    parameter 2"""
    the_diff = coral.Set()
    right_names = right_query.queryelement.name
    for qe in left_query.queryelement:
        if not qe.name in right_names:
            the_diff.append(qe)

    return the_diff

def DPOintersect(left_query, right_query):
    """Find the query elements that have common names"""
    the_common = coral.Set()
    right_names = right_query.queryelement.name
    for qe in left_query.queryelement:
        if qe.name in right_names:
            the_common.append(qe)

    return the_common

class TransformationRules:
    def __init__(self, dpoTrans):
        self.delete = DPOminus(dpoTrans.lhs.query, dpoTrans.rhs)
        self.add = DPOminus(dpoTrans.rhs, dpoTrans.lhs.query)
        self.keep = DPOintersect(dpoTrans.lhs.query, dpoTrans.rhs)
        self.lhs2rhs = self.create_map(self.keep, dpoTrans.rhs.queryelement)
        

        # TODO add pre and post scripts to the metamodel:
        self.pre_executable_code = None
        self.post_executable_code = None

        if dpoTrans.prescript:
           self.pre_executable_code = compile(str(dpoTrans.prescript), "script", "exec")
        if dpoTrans.postscript:
           self.post_executable_code = compile(str(dpoTrans.postscript), "script", "exec")
        
        #print "Calculated rules\n"

    def create_map(self, fro, to):
        theMap = {}
        for elem in fro:
            theMap[elem.abstractElement] = to.select(lambda qe: qe.name == elem.name)[0].abstractElement

        return theMap
    
    def apply(self, match, target):
        
        rhs2lhs_map = swap_key_value(self.lhs2rhs)

        element_propname_items_removed = slot_diff(self.lhs2rhs) # Check for slot differences
        element_propname_items_added = slot_diff(rhs2lhs_map) # Check for slot differences

        # In future it might be a good idea to rewrite the rewrite function, to clean it up.
        rewrite(target, match, self.pre_executable_code, self.post_executable_code, rhs2lhs_map, element_propname_items_removed, element_propname_items_added, self.add.abstractElement, self.delete.abstractElement)

        #print "Applied precalculated rules on a match\n"
        

def DPOTransform(dpoTrans, target):
    """Transform using DPOTRANS10::DPORule"""
    
    rules = TransformationRules(dpoTrans)

    for match in coral.addins.MATCH10.evaluate.GraphMatchEvaluator(dpoTrans.lhs).search(target):
        #unfortunately graphmatch is using names as keys, converting back
        elem_match = {}
        for key, value in match.items():
            elem_match[ dpoTrans.lhs.query.queryelement.select(lambda qe: unicode(qe.name) == key)[0].abstractElement ] = value
        
        rules.apply(elem_match, target)

